# -*- coding: utf-8 -*-
# Copyright: (c) 2021, Jordan Borean (@jborean93) <jborean93@gmail.com>
# MIT License (see LICENSE or https://opensource.org/licenses/MIT)

import functools
import typing
import uuid

from psrpcore._client import (
    ClientGetCommandMetadata,
    ClientPowerShell,
    ClientRunspacePool,
)
from psrpcore._server import ServerPipeline, ServerRunspacePool
from psrpcore.types import (
    BufferCell,
    BufferCellType,
    ChoiceDescription,
    ConsoleColor,
    ControlKeyStates,
    Coordinates,
    FieldDescription,
    HostMethodIdentifier,
    KeyInfo,
    ProgressRecordMsg,
    ProgressRecordType,
    PSChar,
    PSCredential,
    PSCredentialTypes,
    PSCredentialUIOptions,
    PSListBase,
    PSObject,
    PSSecureString,
    PSType,
    PSVersion,
    ReadKeyOptions,
    Rectangle,
    Size,
    add_note_property,
)


@PSType(["System.Int32[]", "System.Array"])
class _ArrayCount(PSListBase):
    """Used to encapsuilate an int32 array for multidimensional arrays."""


def _create_multi_dimensional_array(
    values: typing.List[typing.List[typing.Any]],
    pack_func: typing.Callable[[typing.Any], PSObject],
) -> PSObject:
    """Create a host method multidimensional array object."""
    flattened: typing.List[typing.Any] = []
    for val in values:
        flattened.extend(val)

    # The code only handles 2 dimensional arrays for now.
    mal = _ArrayCount([len(values), len(values[0])])
    mae = [pack_func(v) for v in flattened]

    obj = PSObject()
    obj.PSObject.type_names = []
    add_note_property(obj, "mae", mae)
    add_note_property(obj, "mal", mal)

    return obj


class ClientHostResponder:
    """Client Host Responser.

    Helper class that can be used to response to runspace and pipeline host
    calls. Each method is named after the corresponding host method identifer
    and document the required values that need to be set to respond to the hsot
    call for that request.

    Args:
        connection: The client Runspace Pool or Pipeline to response on.
    """

    def __init__(
        self,
        connection: typing.Union[ClientRunspacePool, ClientPowerShell, ClientGetCommandMetadata],
    ):
        self._connection = connection

    def get_name(
        self,
        ci: int,
        name: str,
    ) -> None:
        """GetName Response.

        Responds to :class:`psrpcore.types.HostMethodIdentifier.GetName` with a
        human friendly host name identifier.

        Note:
            For pwsh this value is typically static and set by the server on
            creation. It is rare for this to be requested by the server.

        Args:
            ci: The call id the response is for.
            name: The host name to respond.
        """
        self._connection.host_response(ci, name)

    def get_version(
        self,
        ci: int,
        version: typing.Union[str, PSVersion],
    ) -> None:
        """GetVersion Response.

        Responds to :class:`psrpcore.types.HostMethodIdentifier.GetVersion`
        with the version of the host.

        Note:
            For pwsh this value is typically static and set by the server on
            creation. It is rare for this to be requested by the server.

        Args:
            ci: The call id the response is for.
            version: The version of the host.
        """
        if not isinstance(version, PSVersion):
            version = PSVersion(version)

        self._connection.host_response(ci, version)

    def get_instance_id(
        self,
        ci: int,
        instance_id: uuid.UUID,
    ) -> None:
        """GetInstanceId Response.

        Responds to :class:`psrpcore.types.HostMethodIdentifier.GetInstanceId`
        with a unique identifier of the host.

        Note:
            For pwsh this value is typically static and set by the server on
            creation. It is rare for this to be requested by the server.

        Args:
            ci: The call id the response is for.
            instance_id: The host identifier.
        """
        self._connection.host_response(ci, instance_id)

    def get_current_culture(
        self,
        ci: int,
        culture: str,
    ) -> None:
        """GetCurrentCulture Response.

        Responds to
        :class:`psrpcore.types.HostMethodIdentifier.GetCurrentCulture` with the
        host's culture, like ``en-US``.

        Note:
            For pwsh this value is typically static and set by the server on
            creation. It is rare for this to be requested by the server.

        Args:
            ci: The call id the response is for.
            culture: The host culture.
        """
        self._connection.host_response(ci, culture)

    def get_current_ui_culture(
        self,
        ci: int,
        culture: str,
    ) -> None:
        """GetCurrentUICulture Response.

        Responds to
        :class:`psrpcore.types.HostMethodIdentifier.GetCurrentUICulture` with
        the host's UI culture, like ``en-US``. The UI culture is used for
        operations like loading resource files.

        Note:
            For pwsh this value is typically static and set by the server on
            creation. It is rare for this to be requested by the server.

        Args:
            ci: The call id the response is for.
            culture: The host UI culture.
        """
        self._connection.host_response(ci, culture)

    def get_is_runspace_pushed(
        self,
        ci: int,
        is_pushed: bool,
    ) -> None:
        """GetIsRunspacePushed Response.

        Responds to
        :class:`psrpcore.types.HostMethodIdentifier.GetIsRunspacePushed` with
        whether the runspace is pushed or not.

        Note:
            For pwsh this value is typically static and set by the server on
            creation. It is rare for this to be requested by the server.

        Args:
            ci: The call id the response is for.
            is_pushed: The runspace is pushed or not.
        """
        self._connection.host_response(ci, is_pushed)

    def read_line(
        self,
        ci: int,
        line: str,
    ) -> None:
        """ReadLine Response.

        Responds to
        :class:`psrpcore.types.HostMethodIdentifier.ReadLine` with the line
        that was read.

        Args:
            ci: The call id the response is for.
            line: The line that was read.
        """
        self._connection.host_response(ci, line)

    def read_line_as_secure_string(
        self,
        ci: int,
        line: PSSecureString,
    ) -> None:
        """ReadLineAsSecureString Response.

        Responds to
        :class:`psrpcore.types.HostMethodIdentifier.ReadLineAsSecureString`
        with the line as a secure string that was read.

        Args:
            ci: The call id the response is for.
            line: The line that was read as a secure string.
        """
        self._connection.host_response(ci, line)

    def prompt(
        self,
        ci: int,
        choices: typing.Dict[str, typing.Any],
    ) -> None:
        """Prompt Response.

        Responds to
        :class:`psrpcore.types.HostMethodIdentifier.Prompt` with the values for
        field that was entered.

        Args:
            ci: The call id the response is for.
            choices: The response with the key being the name of the field and
                the value is the value entered for that field.
        """
        self._connection.host_response(ci, choices)

    def prompt_for_credential(
        self,
        ci: int,
        credential: PSCredential,
    ) -> None:
        """PromptForCredential Response.

        Responds to
        :class:`psrpcore.types.HostMethodIdentifier.PromptForCredential1` or
        :class:`psrpcore.types.HostMethodIdentifier.PromptForCredential2` with
        the credential that was entered.

        Args:
            ci: The call id the response is for.
            credential: The credential that was entered.
        """
        self._connection.host_response(ci, credential)

    def prompt_for_choice(self, ci: int, choice: int) -> None:
        """PromptForChoice Response.

        Responds to
        :class:`psrpcore.types.HostMethodIdentifier.PromptForChoice` with the
        choice that was slected.

        Args:
            ci: The call id the response is for.
            choice: The choice that was selected. This is a 0 based index of
                selected choice from the host call.
        """
        self._connection.host_response(ci, choice)

    def prompt_for_multiple_choice(
        self,
        ci: int,
        choices: typing.Union[int, typing.List[int]],
    ) -> None:
        """PromptForChoiceMultipleSelection Response.

        Responds to
        :class:`psrpcore.types.HostMethodIdentifier.PromptForChoiceMultipleSelection`
        with the choices that were selected.

        Args:
            ci: The call id the response is for.
            choices: The choices that were selected. Each choice is a 0 based
                index of the selected choice from the host call.
        """
        self._connection.host_response(ci, choices if isinstance(choices, list) else [choices])

    def get_foreground_color(
        self,
        ci: int,
        color: ConsoleColor,
    ) -> None:
        """GetForegroundColor Response.

        Responds to
        :class:`psrpcore.types.HostMethodIdentifier.GetForegroundColor` with
        the foreground (text) color of the host.

        Note:
            For pwsh this value is set when the server creates the host based
            on the host default data. It is rare for this to be requested by
            the server.

        Args:
            ci: The call id the response is for.
            color: The foreground color of the host.
        """
        self._connection.host_response(ci, color.value)

    def get_background_color(
        self,
        ci: int,
        color: ConsoleColor,
    ) -> None:
        """GetBackgroundColor Response.

        Responds to
        :class:`psrpcore.types.HostMethodIdentifier.GetBackgroundColor` with
        the background color of the host.

        Note:
            For pwsh this value is set when the server creates the host based
            on the host default data. It is rare for this to be requested by
            the server.

        Args:
            ci: The call id the response is for.
            color: The background color of the host.
        """
        self._connection.host_response(ci, color.value)

    def get_cursor_position(
        self,
        ci: int,
        x: int,
        y: int,
    ) -> None:
        """GetCursorPosition Response.

        Responds to
        :class:`psrpcore.types.HostMethodIdentifier.GetCursorPosition` the
        coordinates of the cursor in the host.

        Note:
            For pwsh this value is set when the server creates the host based
            on the host default data. It is rare for this to be requested by
            the server.

        Args:
            ci: The call id the response is for.
            x: The X (horizontal) coordinate of the cursor.
            y: The Y (vertical) coordinate of the cursor.
        """
        self._connection.host_response(ci, Coordinates.ToPSObjectForRemoting(Coordinates(x, y), for_host=True))

    def get_window_position(
        self,
        ci: int,
        x: int,
        y: int,
    ) -> None:
        """GetWindowPosition Response.

        Responds to
        :class:`psrpcore.types.HostMethodIdentifier.GetWindowPosition` the
        coordinates (from the top left point) of the host window.

        Note:
            For pwsh this value is set when the server creates the host based
            on the host default data. It is rare for this to be requested by
            the server.

        Args:
            ci: The call id the response is for.
            x: The X (horizontal) coordinate of the host window.
            y: The Y (vertical) coordinate of the host window.
        """
        self._connection.host_response(ci, Coordinates.ToPSObjectForRemoting(Coordinates(x, y), for_host=True))

    def get_cursor_size(
        self,
        ci: int,
        size: int,
    ) -> None:
        """GetCursorSize Response.

        Responds to
        :class:`psrpcore.types.HostMethodIdentifier.GetCursorSize` with the
        size (as a percentage) of the cursor.

        Note:
            For pwsh this value is set when the server creates the host based
            on the host default data. It is rare for this to be requested by
            the server.

        Args:
            ci: The call id the response is for.
            size: The size, as a percentage, of the host cursor.
        """
        self._connection.host_response(ci, size)

    def get_buffer_size(
        self,
        ci: int,
        width: int,
        height: int,
    ) -> None:
        """GetBufferSize Response.

        Responds to
        :class:`psrpcore.types.HostMethodIdentifier.GetBufferSize` with the
        size of the buffer of the host. The size is measured in character cells
        which can contain a 16-bit Char.

        Note:
            For pwsh this value is set when the server creates the host based
            on the host default data. It is rare for this to be requested by
            the server.

        Args:
            ci: The call id the response is for.
            width: The number of cells the fit the width of the buffer.
            height: The number of cells that fit the height of the buffer.
        """
        self._connection.host_response(ci, Size.ToPSObjectForRemoting(Size(width, height), for_host=True))

    def get_window_size(
        self,
        ci: int,
        width: int,
        height: int,
    ) -> None:
        """GetWindowSize Response.

        Responds to
        :class:`psrpcore.types.HostMethodIdentifier.GetWindowSize` with the
        size of the window of the host. The size is measured in character cells
        which can contain a 16-bit Char.

        Note:
            For pwsh this value is set when the server creates the host based
            on the host default data. It is rare for this to be requested by
            the server.

        Args:
            ci: The call id the response is for.
            width: The number of cells the fit the width of the window.
            height: The number of cells that fit the height of the window.
        """
        self._connection.host_response(ci, Size.ToPSObjectForRemoting(Size(width, height), for_host=True))

    def get_window_title(
        self,
        ci: int,
        title: str,
    ) -> None:
        """GetWindowTitle Response.

        Responds to
        :class:`psrpcore.types.HostMethodIdentifier.GetWindowTitle` with the
        title of the host window.

        Note:
            For pwsh this value is set when the server creates the host based
            on the host default data. It is rare for this to be requested by
            the server.

        Args:
            ci: The call id the response is for.
            title: The title of the window.
        """
        self._connection.host_response(ci, title)

    def get_max_window_size(
        self,
        ci: int,
        width: int,
        height: int,
    ) -> None:
        """GetMaxWindowSize Response.

        Responds to
        :class:`psrpcore.types.HostMethodIdentifier.GetMaxWindowSize` with the
        maximum size of the window of the host. The size is measured in
        character cells which can contain a 16-bit Char.

        Note:
            For pwsh this value is set when the server creates the host based
            on the host default data. It is rare for this to be requested by
            the server.

        Args:
            ci: The call id the response is for.
            width: The number of cells the fit the max width of the window.
            height: The number of cells that fit the max height of the window.
        """
        self._connection.host_response(ci, Size.ToPSObjectForRemoting(Size(width, height), for_host=True))

    def get_max_physical_window_size(
        self,
        ci: int,
        width: int,
        height: int,
    ) -> None:
        """GetMaxPhysicalWindowSize Response.

        Responds to
        :class:`psrpcore.types.HostMethodIdentifier.GetMaxPhysicalWindowSize`
        with the maximum physical size of the window of the host. The size is
        measured in character cells which can contain a 16-bit Char.

        Note:
            For pwsh this value is set when the server creates the host based
            on the host default data. It is rare for this to be requested by
            the server.

        Args:
            ci: The call id the response is for.
            width: The number of cells the fit the max physical width of the
                window.
            height: The number of cells that fit the max physical height of
                the window.
        """
        self._connection.host_response(ci, Size.ToPSObjectForRemoting(Size(width, height), for_host=True))

    def get_key_available(
        self,
        ci: int,
        waiting: bool,
    ) -> None:
        """GetKeyAvailable Response.

        Responds to
        :class:`psrpcore.types.HostMethodIdentifier.GetKeyAvailable` with
        whether the host is waiting on input.

        Args:
            ci: The call id the response is for.
            waiting: Host is waiting on input.
        """
        self._connection.host_response(ci, waiting)

    def read_key(
        self,
        ci: int,
        character: typing.Union[int, str, PSChar],
        key_down: bool,
        control_key_state: ControlKeyStates = ControlKeyStates.none,
        key_code: int = 0,
    ) -> None:
        """ReadKey Response.

        Responds to
        :class:`psrpcore.types.HostMethodIdentifier.ReadKey` with key that was
        inputted onto the host.

        Args:
            ci: The call id the response is for.
            character: The character (UTF-16 16-bit char) that was inputted.
            key_down: Generated by a key press ``True`` or release ``False``.
            control_key_state: State of the control keys.
            key_code: Device-independent key code value.
        """
        char = character
        if not isinstance(char, PSChar):
            char = PSChar(character)

        self._connection.host_response(
            ci, KeyInfo.ToPSObjectForRemoting(KeyInfo(key_code, char, control_key_state, key_down), for_host=True)
        )

    def get_buffer_contents(
        self,
        ci: int,
        cells: typing.List[typing.List[BufferCell]],
    ) -> None:
        """GetBufferContents Response.

        Responds to
        :class:`psrpcore.types.HostMethodIdentifier.GetBufferContents` with
        contents of the buffer that was requested.

        The ``cells`` value is a list of a list of
        :class:`psrpcore.types.BufferCell` where the first list dimension
        represents each row and the 2nd dimension is each column of that row.

        Note:
            PowerShell does not implement this call for security reasons.
            Unless you trust the target host you should not return this
            information, or sanitise the data as needed.

        Args:
            ci: The call id the response is for.
            cells: A list of a list of cells.
        """
        wrap_cell = functools.partial(BufferCell.ToPSObjectForRemoting, for_host=True)
        return self._connection.host_response(ci, _create_multi_dimensional_array(cells, wrap_cell))


class ServerHostRequestor:
    """Server Host Requestor.

    Helper class that can be used to create Runspace Pool or Pipeline host
    call requests. Each method exposes the required/optional arguments for each
    call and extra handling that may be required for that request. The
    :class:`ClientHostResponder` is used on the client side to respond to the
    host calls that require a response.

    Args:
        connection: The server Runspace Pool or Pipeline to send requests to.
    """

    def __init__(
        self,
        connection: typing.Union[ServerRunspacePool, ServerPipeline],
    ):
        self._connection = connection

    def get_name(
        self,
    ) -> int:
        """GetName Request.

        Sends the :class:`psrpcore.types.HostMethodIdentifier.GetName` host
        call that requests the human friendly host name identifier.
        human friendly host name identifier.

        This corresponds to the `PSHost.Name Property`_.

        Note:
            The server should wait for the host response before continuing the
            pipeline that created this request.

        Returns:
            int: The call id for the request.

        .. _PSHost.Name Property:
            https://docs.microsoft.com/en-us/dotnet/api/system.management.automation.host.pshost.name
        """
        return self._connection.host_call(HostMethodIdentifier.GetName)

    def get_version(
        self,
    ) -> int:
        """GetVersion Request.

        Sends the :class:`psrpcore.types.HostMethodIdentifier.GetVersion` host
        call that requests the version of the hosting application.

        This corresponds to the `PSHost.Version Property`_.

        Note:
            The server should wait for the host response before continuing the
            pipeline that created this request.

        Returns:
            int: The call id for the request.

        .. _PSHost.Version Property:
            https://docs.microsoft.com/en-us/dotnet/api/system.management.automation.host.pshost.version
        """
        return self._connection.host_call(HostMethodIdentifier.GetVersion)

    def get_instance_id(
        self,
    ) -> int:
        """GetInstanceId Request.

        Sends the :class:`psrpcore.types.HostMethodIdentifier.GetInstanceId`
        host call that requests the identifier of the hosting application.

        This corresponds to the `PSHost.InstanceId Property`_.

        Note:
            The server should wait for the host response before continuing the
            pipeline that created this request.

        Returns:
            int: The call id for the request.

        .. _PSHost.InstanceId Property:
            https://docs.microsoft.com/en-us/dotnet/api/system.management.automation.host.pshost.instanceid
        """
        return self._connection.host_call(HostMethodIdentifier.GetInstanceId)

    def get_current_culture(
        self,
    ) -> int:
        """GetCurrentCulture Request.

        Sends the
        :class:`psrpcore.types.HostMethodIdentifier.GetCurrentCulture` host
        call that requests the current culture of the host.

        This corresponds to the `PSHost.CurrentCulture Property`_.

        Note:
            The server should wait for the host response before continuing the
            pipeline that created this request.

        Returns:
            int: The call id for the request.

        .. _PSHost.CurrentCulture Property:
            https://docs.microsoft.com/en-us/dotnet/api/system.management.automation.host.pshost.currentculture
        """
        return self._connection.host_call(HostMethodIdentifier.GetCurrentCulture)

    def get_current_ui_culture(
        self,
    ) -> int:
        """GetCurrentUICulture Request.

        Sends the
        :class:`psrpcore.types.HostMethodIdentifier.GetCurrentUICulture` host
        call that requests the current UI culture of the host.

        This corresponds to the `PSHost.CurrentUICulture Property`_.

        Note:
            The server should wait for the host response before continuing the
            pipeline that created this request.

        Returns:
            int: The call id for the request.

        .. _PSHost.CurrentUICulture Property:
            https://docs.microsoft.com/en-us/dotnet/api/system.management.automation.host.pshost.currentuiculture
        """
        return self._connection.host_call(HostMethodIdentifier.GetCurrentUICulture)

    def set_should_exit(
        self,
        exit_code: int,
    ) -> None:
        """SetShouldExit Request.

        Sends the
        :class:`psrpcore.types.HostMethodIdentifier.SetShouldExit` host
        call that requests the current engine runspace to shut down and
        terminate the host's root runspace.

        This corresponds to the `PSHost.SetShouldExit Method`_.

        Note:
            This is a void method and the server should continue pipeline
            execution and expect no response from the client.

        Args:
            exit_code: The exit code accompanying the exit keyword.

        .. _PSHost.SetShouldExit Method:
            https://docs.microsoft.com/en-us/dotnet/api/system.management.automation.host.pshost.setshouldexit
        """
        self._connection.host_call(HostMethodIdentifier.SetShouldExit, [exit_code])

    def enter_nested_prompt(
        self,
    ) -> None:
        """EnterNestedPrompt Request.

        Sends the
        :class:`psrpcore.types.HostMethodIdentifier.EnterNestedPrompt` host
        call that instructs the host to interrupt the currently running
        pipeline and start a new, "nested" input loop, where an input loop is
        the cycle of prompt, input, execute.

        This corresponds to the `PSHost.EnterNestedPrompt Method`_.

        Note:
            This is a void method and the server should continue pipeline
            execution and expect no response from the client.

        .. _PSHost.EnterNestedPrompt Method:
            https://docs.microsoft.com/en-us/dotnet/api/system.management.automation.host.pshost.enternestedprompt
        """
        self._connection.host_call(HostMethodIdentifier.EnterNestedPrompt)

    def exit_nested_prompt(
        self,
    ) -> None:
        """ExitNestedPrompt Request.

        Sends the
        :class:`psrpcore.types.HostMethodIdentifier.ExitNestedPrompt` host
        call that causes the host to end the currently running input loop. If
        the input loop was created by a prior call to EnterNestedPrompt, the
        enclosing pipeline will be resumed. If the current input loop is the
        top-most loop, then the host will act as though
        :meth:`set_should_exit()` was called.

        This corresponds to the `PSHost.ExitNestedPrompt Method`_.

        Note:
            This is a void method and the server should continue pipeline
            execution and expect no response from the client.

        .. _PSHost.ExitNestedPrompt Method:
            https://docs.microsoft.com/en-us/dotnet/api/system.management.automation.host.pshost.exitnestedprompt
        """
        self._connection.host_call(HostMethodIdentifier.ExitNestedPrompt)

    def notify_begin_application(
        self,
    ) -> None:
        """NotifyBeginApplication Request.

        Sends the
        :class:`psrpcore.types.HostMethodIdentifier.NotifyBeginApplication`
        host call that is called by the engine to notify the host that it is
        about to execute a "legacy" command line application. A legacy
        application is defined as a console-mode executable that may do one or
        more of the following:

            * reads from stdin
            * writes to stdout
            * writes to stderr

        The engine will always call this method and
        :meth:`notify_end_application()` in pairs.

        This corresponds to the `PSHost.NotifyBeginApplication Method`_.

        Note:
            This is a void method and the server should continue pipeline
            execution and expect no response from the client.

        .. _PSHost.NotifyBeginApplication Method:
            https://docs.microsoft.com/en-us/dotnet/api/system.management.automation.host.pshost.notifybeginapplication
        """
        self._connection.host_call(HostMethodIdentifier.NotifyBeginApplication)

    def notify_end_application(
        self,
    ) -> None:
        """NotifyEndApplication Request.

        Sends the
        :class:`psrpcore.types.HostMethodIdentifier.NotifyEndApplication`
        host call that is called by the engine to notify the host that the
        execution of a legacy command has completed.

        The engine will always call this method and
        :meth:`notify_begin_application()` in pairs.

        This corresponds to the `PSHost.NotifyEndApplication Method`_.

        Note:
            This is a void method and the server should continue pipeline
            execution and expect no response from the client.

        .. _PSHost.NotifyEndApplication Method:
            https://docs.microsoft.com/en-us/dotnet/api/system.management.automation.host.pshost.notifyendapplication
        """
        self._connection.host_call(HostMethodIdentifier.NotifyEndApplication)

    # I don't think this is actually possible to do.
    # def push_runspace(self, runspace) -> None:
    #     self._connection.host_call(HostMethodIdentifier.PushRunspace, [runspace])

    def pop_runspace(
        self,
    ) -> None:
        """PopRunspace Request.

        Sends the :class:`psrpcore.types.HostMethodIdentifier.PopRunspace` host
        call that is called by the engine to notify the host that a Runspace
        pop has been requested.

        This corresponds to the `IHostSupportsInteractiveSession.PopRunspace Method`_.

        Note:
            This is a void method and the server should continue pipeline
            execution and expect no response from the client.

        .. _IHostSupportsInteractiveSession.PopRunspace Method:
            https://docs.microsoft.com/en-us/dotnet/api/system.management.automation.host.ihostsupportsinteractivesession.poprunspace
        """
        self._connection.host_call(HostMethodIdentifier.PopRunspace)

    def get_is_runspace_pushed(
        self,
    ) -> int:
        """GetIsRunspacePushed Request.

        Sends the
        :class:`psrpcore.types.HostMethodIdentifier.GetIsRunspacePushed` host
        call to check if the runspace is pushed or not.

        This corresponds to the `IHostSupportsInteractiveSession.IsRunspacePushed Property`_.

        Note:
            The server should wait for the host response before continuing the
            pipeline that created this request.

        Returns:
            int: The call id for the request.

        .. _IHostSupportsInteractiveSession.IsRunspacePushed Property:
            https://docs.microsoft.com/en-us/dotnet/api/system.management.automation.host.ihostsupportsinteractivesession.isrunspacepushed
        """
        return self._connection.host_call(HostMethodIdentifier.GetIsRunspacePushed)

    # Also don't think this is possible remotely.
    # def get_runspace(self) -> int:
    #     return self._connection.host_call(HostMethodIdentifier.GetRunspace, [])

    def read_line(
        self,
    ) -> int:
        """ReadLine Request.

        Sends the :class:`psrpcore.types.HostMethodIdentifier.ReadLine` host
        call to read characters from the console until a newline is
        encountered.

        This corresponds to the `PSHostUserInterface.ReadLine Method`_.

        Note:
            The server should wait for the host response before continuing the
            pipeline that created this request.

        Returns:
            int: The call id for the request.

        .. _PSHostUserInterface.ReadLine Method:
            https://docs.microsoft.com/en-us/dotnet/api/system.management.automation.host.pshostuserinterface.readline
        """
        return self._connection.host_call(HostMethodIdentifier.ReadLine)

    def read_line_as_secure_string(
        self,
    ) -> int:
        """ReadLineAsSecureString Request.

        Sends the :class:`psrpcore.types.HostMethodIdentifier.ReadLineAsSecureString` host
        call to read characters from the console until a newline is
        encountered without echoing the input back to the user.

        This corresponds to the `PSHostUserInterface.ReadLineAsSecureString Method`_.

        Note:
            The server should wait for the host response before continuing the
            pipeline that created this request.

        Returns:
            int: The call id for the request.

        .. _PSHostUserInterface.ReadLineAsSecureString Method:
            https://docs.microsoft.com/en-us/dotnet/api/system.management.automation.host.pshostuserinterface.readlineassecurestring
        """
        return self._connection.host_call(HostMethodIdentifier.ReadLineAsSecureString)

    def write(
        self,
        value: str,
        foreground_color: typing.Optional[ConsoleColor] = None,
        background_color: typing.Optional[ConsoleColor] = None,
    ) -> None:
        """Write Request.

        Sends the :class:`psrpcore.types.HostMethodIdentifier.Write1` or
        :class:`psrpcore.types.HostMethodIdentifier.Write2` host call that
        writes the characters to the screen buffer.

        This corresponds to the `PSHostUserInterface.Write Method`_.

        Note:
            This is a void method and the server should continue pipeline
            execution and expect no response from the client.

        Args:
            value: The characters to write.
            foreground_color: The color to display the text with.
            background_color: The color to display the background with.

        .. _PSHostUserInterface.Write Method:
            https://docs.microsoft.com/en-us/dotnet/api/system.management.automation.host.pshostuserinterface.write
        """
        mi = HostMethodIdentifier.Write1
        mp: typing.List[typing.Union[str, int]] = [value]

        if not (foreground_color is None and background_color is None):
            mi = HostMethodIdentifier.Write2
            mp.insert(0, foreground_color.value if foreground_color is not None else 0)
            mp.insert(1, background_color.value if background_color is not None else 0)

        self._connection.host_call(mi, mp)

    def write_line(
        self,
        line: typing.Optional[str] = None,
        foreground_color: typing.Optional[ConsoleColor] = None,
        background_color: typing.Optional[ConsoleColor] = None,
    ) -> None:
        """WriteLine Request.

        Sends the :class:`psrpcore.types.HostMethodIdentifier.WriteLine1`,
        :class:`psrpcore.types.HostMethodIdentifier.WriteLine2`, or
        :class:`psrpcore.types.HostMethodIdentifier.WriteLine2` host call that
        writes the characters with a newline to the screen buffer.

        This corresponds to the `PSHostUserInterface.WriteLine Method`_.

        Note:
            This is a void method and the server should continue pipeline
            execution and expect no response from the client.

        Args:
            line: The line to write, if not set then just a newline is written.
            foreground_color: The color to display the line with.
            background_color: The color to display the background with.

        .. _PSHostUserInterface.WriteLine Method:
            https://docs.microsoft.com/en-us/dotnet/api/system.management.automation.host.pshostuserinterface.writeline
        """
        mi = HostMethodIdentifier.WriteLine1
        mp: typing.List[typing.Union[str, int]] = []

        if line is not None:
            mi = HostMethodIdentifier.WriteLine2
            mp.append(line)

        if not (foreground_color is None and background_color is None):
            if not mp:
                mp.append("")

            mi = HostMethodIdentifier.WriteLine3
            mp.insert(0, foreground_color.value if foreground_color is not None else 0)
            mp.insert(1, background_color.value if background_color is not None else 0)

        self._connection.host_call(mi, mp)

    def write_error_line(
        self,
        line: str,
    ) -> None:
        """WriteErrorLine Request.

        Sends the :class:`psrpcore.types.HostMethodIdentifier.WriteErrorLine`
        host call that writes the line to the error display of the host.

        This corresponds to the `PSHostUserInterface.WriteErrorLine Method`_.

        Note:
            This is a void method and the server should continue pipeline
            execution and expect no response from the client.

        Args:
            line: The line to write to the error display.

        .. _PSHostUserInterface.WriteErrorLine Method:
            https://docs.microsoft.com/en-us/dotnet/api/system.management.automation.host.pshostuserinterface.writeerrorline
        """
        self._connection.host_call(HostMethodIdentifier.WriteErrorLine, [line])

    def write_debug_line(
        self,
        line: str,
    ) -> None:
        """WriteDebugLine Request.

        Sends the :class:`psrpcore.types.HostMethodIdentifier.WriteDebugLine`
        host call that writes a debugging message to the host.

        This corresponds to the `PSHostUserInterface.WriteDebugLine Method`_.

        Note:
            This is a void method and the server should continue pipeline
            execution and expect no response from the client.

        Args:
            line: The debug line to write to the display.

        .. _PSHostUserInterface.WriteDebugLine Method:
            https://docs.microsoft.com/en-us/dotnet/api/system.management.automation.host.pshostuserinterface.writedebugline
        """
        self._connection.host_call(HostMethodIdentifier.WriteDebugLine, [line])

    def write_progress(
        self,
        source_id: int,
        activity_id: int,
        activity: str,
        status_description: str,
        current_operation: typing.Optional[str] = None,
        parent_activity_id: int = -1,
        percent_complete: int = -1,
        record_type: ProgressRecordType = ProgressRecordType.Processing,
        seconds_remaining: int = -1,
    ) -> None:
        """WriteProgress Request.

        Sends the :class:`psrpcore.types.HostMethodIdentifier.WriteProgress`
        host call that writes a progress record to be displayed on the host.

        This corresponds to the `PSHostUserInterface.WriteProgress Method`_.

        Note:
            This is a void method and the server should continue pipeline
            execution and expect no response from the client.

        Args:
            source_id: Unique identifier of the source of the record.
            activity: The description of the activity for which progress is
                being reported.
            activity_id: The Id of the activity to which this record
                corresponds. Used as a key for linking of subordinate
                activities.
            status_description: Current status of the operation, e.g.
                "35 of 50 items copied.".
            current_operation: Current operation of the many required to
                accomplish the activity, e.g. "copying foo.txt".
            parent_activity_id: The Id of the activity for which this record is
                a subordinate.
            percent_complete: The estimate of the percentage of total work for
                the activity that is completed. Set to a negative value to
                indicate that the percentage completed should not be displayed.
            record_type: The type of record represented.
            seconds_remaining: The estimate of time remaining until this
                activity is completed. Set to a negative value to indicate that
                the seconds remaining should not be displayed.

        .. _PSHostUserInterface.WriteProgress Method:
            https://docs.microsoft.com/en-us/dotnet/api/system.management.automation.host.pshostuserinterface.writeprogress
        """
        # ProgressRecord serialized here is like a PSRP ProgressRecordMsg object.
        self._connection.host_call(
            HostMethodIdentifier.WriteProgress,
            [
                source_id,
                ProgressRecordMsg(
                    ActivityId=activity_id,
                    Activity=activity,
                    StatusDescription=status_description,
                    CurrentOperation=current_operation,
                    ParentActivityId=parent_activity_id,
                    PercentComplete=percent_complete,
                    Type=record_type,
                    SecondsRemaining=seconds_remaining,
                ),
            ],
        )

    def write_verbose_line(
        self,
        line: str,
    ) -> None:
        """WriteVerboseLine Request.

        Sends the :class:`psrpcore.types.HostMethodIdentifier.WriteVerboseLine`
        host call that writes a verbose message to the host.

        This corresponds to the `PSHostUserInterface.WriteVerboseLine Method`_.

        Note:
            This is a void method and the server should continue pipeline
            execution and expect no response from the client.

        Args:
            line: The verbose line to write to the display.

        .. _PSHostUserInterface.WriteVerboseLine Method:
            https://docs.microsoft.com/en-us/dotnet/api/system.management.automation.host.pshostuserinterface.writeverboseline
        """
        self._connection.host_call(HostMethodIdentifier.WriteVerboseLine, [line])

    def write_warning_line(
        self,
        line: str,
    ) -> None:
        """WriteWarningLine Request.

        Sends the :class:`psrpcore.types.HostMethodIdentifier.WriteWarningLine`
        host call that writes a warning message to the host.

        This corresponds to the `PSHostUserInterface.WriteWarningLine Method`_.

        Note:
            This is a void method and the server should continue pipeline
            execution and expect no response from the client.

        Args:
            line: The warning line to write to the display.

        .. _PSHostUserInterface.WriteWarningLine Method:
            https://docs.microsoft.com/en-us/dotnet/api/system.management.automation.host.pshostuserinterface.writewarningline
        """
        self._connection.host_call(HostMethodIdentifier.WriteWarningLine, [line])

    def prompt(
        self,
        caption: str,
        message: str,
        descriptions: typing.List[FieldDescription],
    ) -> int:
        """Prompt Request.

        Sends the :class:`psrpcore.types.HostMethodIdentifier.Prompt`
        host call prompts the user with a number of fields for which to supply
        values.

        This corresponds to the `PSHostUserInterface.Prompt Method`_.

        Note:
            The server should wait for the host response before continuing the
            pipeline that created this request.

        Args:
            caption: The caption or title for the prompt.
            message: The message describing the set of fields.
            descriptions: A list of fields to display.

        Returns:
            int: The call id for the request.

        .. _PSHostUserInterface.Prompt Method:
            https://docs.microsoft.com/en-us/dotnet/api/system.management.automation.host.pshostuserinterface.prompt
        """
        fields = [FieldDescription.ToPSObjectForRemoting(f, for_host=True) for f in descriptions]

        return self._connection.host_call(HostMethodIdentifier.Prompt, [caption, message, fields])

    def prompt_for_credential(
        self,
        caption: str = "PSRPCore credential request",
        message: str = "Enter your credentials.",
        username: typing.Optional[str] = None,
        target_name: typing.Optional[str] = None,
        allowed_credential_types: typing.Optional[PSCredentialTypes] = None,
        options: typing.Optional[PSCredentialUIOptions] = None,
    ) -> int:
        """PromptForCredential Request.

        Sends the
        :class:`psrpcore.types.HostMethodIdentifier.PromptForCredential1`, or
        :class:`psrpcore.types.HostMethodIdentifier.PromptForCredential2` host
        call prompts the user for a credential.

        This corresponds to the `PSHostUserInterface.PromptForCredential Method`_.

        Note:
            The server should wait for the host response before continuing the
            pipeline that created this request.

        Args:
            caption: The caption or title for the prompt.
            message: The message describing the credential required.
            username: The username the credential is for, if omitted or
                None/empty the username is requested in the prompt.
            target_name: The domain part of the username if set.
            allowed_credential_types: The types of credential that is being
                requested.
            options: Options to control the UI behaviour.

        Returns:
            int: The call id for the request.

        .. _PSHostUserInterface.PromptForCredential Method:
            https://docs.microsoft.com/en-us/dotnet/api/system.management.automation.host.pshostuserinterface.promptforcredential
        """
        mi = HostMethodIdentifier.PromptForCredential1
        mp: typing.List[typing.Any] = [caption, message, username, target_name or ""]
        if not (allowed_credential_types is None and options is None):
            mi = HostMethodIdentifier.PromptForCredential2
            mp.append(
                PSCredentialTypes.Default.value if allowed_credential_types is None else allowed_credential_types.value
            )
            mp.append(PSCredentialUIOptions.Default.value if options is None else options.value)

        return self._connection.host_call(mi, mp)

    def prompt_for_choice(
        self,
        caption: str,
        message: str,
        choices: typing.List[ChoiceDescription],
        default_choice: int = -1,
    ) -> int:
        """PromptForChoice Request.

        Sends the :class:`psrpcore.types.HostMethodIdentifier.PromptForChoice`
        host call prompts the user to choose an option from a set list.

        This corresponds to the `PSHostUserInterface.PromptForChoice Method`_.
        See :meth:`prompt_for_multiple_choice()` for a way to allow the user to
        select multiple choices.

        Note:
            The server should wait for the host response before continuing the
            pipeline that created this request.

        Args:
            caption: The caption or title for the prompt.
            message: The message describing the set of choices.
            choices: A list of choices.
            default_choice: The default choice that correspond to the index of
                choices, ``-1`` means no default.

        Returns:
            int: The call id for the request.

        .. _PSHostUserInterface.PromptForChoice Method:
            https://docs.microsoft.com/en-us/dotnet/api/system.management.automation.host.pshostuserinterface.promptforchoice
        """
        converted_choices = [ChoiceDescription.ToPSObjectForRemoting(c, for_host=True) for c in choices]

        return self._connection.host_call(
            HostMethodIdentifier.PromptForChoice, [caption, message, converted_choices, default_choice]
        )

    def prompt_for_multiple_choice(
        self,
        caption: str,
        message: str,
        choices: typing.List[ChoiceDescription],
        default_choices: typing.Optional[typing.List[int]] = None,
    ) -> int:
        """PromptForChoiceMultipleSelection Request.

        Sends the
        :class:`psrpcore.types.HostMethodIdentifier.PromptForChoiceMultipleSelection`
        host call prompts the user to choose multiple options from a set list.

        This corresponds to the
        `IHostUISupportsMultipleChoiceSelection.PromptForChoice Method`_.

        Note:
            The server should wait for the host response before continuing the
            pipeline that created this request.

        Args:
            caption: The caption or title for the prompt.
            message: The message describing the set of choices.
            choices: A list of choices.
            default_choices: A list of default choice that correspond to the index of
                choices.

        Returns:
            int: The call id for the request.

        .. _IHostUISupportsMultipleChoiceSelection.PromptForChoice Method:
            https://docs.microsoft.com/en-us/dotnet/api/system.management.automation.host.ihostuisupportsmultiplechoiceselection.promptforchoice
        """
        converted_choices = [ChoiceDescription.ToPSObjectForRemoting(c, for_host=True) for c in choices]

        return self._connection.host_call(
            HostMethodIdentifier.PromptForChoiceMultipleSelection,
            [caption, message, converted_choices, default_choices or []],
        )

    def get_foreground_color(
        self,
    ) -> int:
        """GetForegroundColor Request.

        Sends the
        :class:`psrpcore.types.HostMethodIdentifier.GetForegroundColor` host
        call to request the foreground color of the host.

        This corresponds to the
        `PSHostRawUserInterface.ForegroundColor Property`_.

        Note:
            The server should wait for the host response before continuing the
            pipeline that created this request.

        Returns:
            int: The call id for the request.

        .. _PSHostRawUserInterface.ForegroundColor Property:
            https://docs.microsoft.com/en-us/dotnet/api/system.management.automation.host.pshostrawuserinterface.foregroundcolor
        """
        return self._connection.host_call(HostMethodIdentifier.GetForegroundColor)

    def set_foreground_color(
        self,
        color: ConsoleColor,
    ) -> None:
        """SetForegroundColor Request.

        Sends the
        :class:`psrpcore.types.HostMethodIdentifier.SetForegroundColor` host
        call to change the foreground color of the host.

        This corresponds to the
        `PSHostRawUserInterface.ForegroundColor Property`_.

        Note:
            This is a void method and the server should continue pipeline
            execution and expect no response from the client.

        Args:
            color: The color to set the foreground color to.

        .. _PSHostRawUserInterface.ForegroundColor Property:
            https://docs.microsoft.com/en-us/dotnet/api/system.management.automation.host.pshostrawuserinterface.foregroundcolor
        """
        self._connection.host_call(HostMethodIdentifier.SetForegroundColor, [color.value])

    def get_background_color(
        self,
    ) -> int:
        """GetBackgroundColor Request.

        Sends the
        :class:`psrpcore.types.HostMethodIdentifier.GetBackgroundColor` host
        call to request the background color of the host.

        This corresponds to the
        `PSHostRawUserInterface.BackgroundColor Property`_.

        Note:
            The server should wait for the host response before continuing the
            pipeline that created this request.

        Returns:
            int: The call id for the request.

        .. _PSHostRawUserInterface.BackgroundColor Property:
            https://docs.microsoft.com/en-us/dotnet/api/system.management.automation.host.pshostrawuserinterface.backgroundcolor
        """
        return self._connection.host_call(HostMethodIdentifier.GetBackgroundColor)

    def set_background_color(
        self,
        color: ConsoleColor,
    ) -> None:
        """SetBackgroundColor Request.

        Sends the
        :class:`psrpcore.types.HostMethodIdentifier.SetBackgroundColor` host
        call to change the background color of the host.

        This corresponds to the
        `PSHostRawUserInterface.BackgroundColor Property`_.

        Note:
            This is a void method and the server should continue pipeline
            execution and expect no response from the client.

        Args:
            color: The color to set the background color to.

        .. _PSHostRawUserInterface.BackgroundColor Property:
            https://docs.microsoft.com/en-us/dotnet/api/system.management.automation.host.pshostrawuserinterface.backgroundcolor
        """
        self._connection.host_call(HostMethodIdentifier.SetBackgroundColor, [color.value])

    def get_cursor_position(
        self,
    ) -> int:
        """GetCursorPosition Request.

        Sends the
        :class:`psrpcore.types.HostMethodIdentifier.GetCursorPosition` host
        call to get the position of the cursor on the screen buffer.

        This corresponds to the
        `PSHostRawUserInterface.CursorPosition Property`_.

        Note:
            The server should wait for the host response before continuing the
            pipeline that created this request.

        Returns:
            int: The call id for the request.

        .. _PSHostRawUserInterface.CursorPosition Property:
            https://docs.microsoft.com/en-us/dotnet/api/system.management.automation.host.pshostrawuserinterface.cursorposition
        """
        return self._connection.host_call(HostMethodIdentifier.GetCursorPosition)

    def set_cursor_position(
        self,
        x: int,
        y: int,
    ) -> None:
        """SetCursorPosition Request.

        Sends the
        :class:`psrpcore.types.HostMethodIdentifier.SetCursorPosition` host
        call to change the position of the cursor on the screen buffer.

        This corresponds to the
        `PSHostRawUserInterface.CursorPosition Property`_.

        Note:
            This is a void method and the server should continue pipeline
            execution and expect no response from the client.

        Args:
            x: The horizontal location to set the cursor to.
            y: The vertical location to set the cursor to.

        .. _PSHostRawUserInterface.CursorPosition Property:
            https://docs.microsoft.com/en-us/dotnet/api/system.management.automation.host.pshostrawuserinterface.cursorposition
        """
        self._connection.host_call(
            HostMethodIdentifier.SetCursorPosition,
            [Coordinates.ToPSObjectForRemoting(Coordinates(x, y), for_host=True)],
        )

    def get_window_position(
        self,
    ) -> int:
        """GetWindowPosition Request.

        Sends the
        :class:`psrpcore.types.HostMethodIdentifier.GetWindowPosition` host
        call to get the position of the view window relative to the screen
        buffer.

        This corresponds to the
        `PSHostRawUserInterface.WindowPosition Property`_.

        Note:
            The server should wait for the host response before continuing the
            pipeline that created this request.

        Returns:
            int: The call id for the request.

        .. _PSHostRawUserInterface.WindowPosition Property:
            https://docs.microsoft.com/en-us/dotnet/api/system.management.automation.host.pshostrawuserinterface.windowposition
        """
        return self._connection.host_call(HostMethodIdentifier.GetWindowPosition)

    def set_window_position(
        self,
        x: int,
        y: int,
    ) -> None:
        """SetWindowPosition Request.

        Sends the
        :class:`psrpcore.types.HostMethodIdentifier.SetWindowPosition` host
        call to change the position of the view window relative to the screen
        buffer.

        This corresponds to the
        `PSHostRawUserInterface.WindowPosition Property`_.

        Note:
            This is a void method and the server should continue pipeline
            execution and expect no response from the client.

        Args:
            x: The horizontal location to set the window to.
            y: The vertical location to set the window to.

        .. _PSHostRawUserInterface.WindowPosition Property:
            https://docs.microsoft.com/en-us/dotnet/api/system.management.automation.host.pshostrawuserinterface.windowposition
        """
        self._connection.host_call(
            HostMethodIdentifier.SetWindowPosition,
            [Coordinates.ToPSObjectForRemoting(Coordinates(x, y), for_host=True)],
        )

    def get_cursor_size(
        self,
    ) -> int:
        """GetCursorSize Request.

        Sends the
        :class:`psrpcore.types.HostMethodIdentifier.GetCursorSize` host
        call to get the size of the cursor as a percentage.

        This corresponds to the
        `PSHostRawUserInterface.CursorSize Property`_.

        Note:
            The server should wait for the host response before continuing the
            pipeline that created this request.

        Returns:
            int: The call id for the request.

        .. _PSHostRawUserInterface.CursorSize Property:
            https://docs.microsoft.com/en-us/dotnet/api/system.management.automation.host.pshostrawuserinterface.cursorsize
        """
        return self._connection.host_call(HostMethodIdentifier.GetCursorSize)

    def set_cursor_size(
        self,
        size: int,
    ) -> None:
        """SetCursorSize Request.

        Sends the
        :class:`psrpcore.types.HostMethodIdentifier.SetCursorSize` host
        call to set the size of the cursor as a percentage.

        This corresponds to the
        `PSHostRawUserInterface.CursorSize Property`_.

        Note:
            This is a void method and the server should continue pipeline
            execution and expect no response from the client.

        Args:
            size: The cursor size as a percentage.

        .. _PSHostRawUserInterface.CursorSize Property:
            https://docs.microsoft.com/en-us/dotnet/api/system.management.automation.host.pshostrawuserinterface.cursorsize
        """
        self._connection.host_call(HostMethodIdentifier.SetCursorSize, [size])

    def get_buffer_size(
        self,
    ) -> int:
        """GetBufferSize Request.

        Sends the
        :class:`psrpcore.types.HostMethodIdentifier.GetBufferSize` host
        call to get the size of the screen buffer as measured by cells.

        This corresponds to the
        `PSHostRawUserInterface.BufferSize Property`_.

        Note:
            The server should wait for the host response before continuing the
            pipeline that created this request.

        Returns:
            int: The call id for the request.

        .. _PSHostRawUserInterface.BufferSize Property:
            https://docs.microsoft.com/en-us/dotnet/api/system.management.automation.host.pshostrawuserinterface.buffersize
        """
        return self._connection.host_call(HostMethodIdentifier.GetBufferSize)

    def set_buffer_size(
        self,
        width: int,
        height: int,
    ) -> None:
        """SetBufferSize Request.

        Sends the
        :class:`psrpcore.types.HostMethodIdentifier.SetBufferSize` host
        call to change the size of the screen buffer as measured by cells.

        This corresponds to the
        `PSHostRawUserInterface.BufferSize Property`_.

        Note:
            This is a void method and the server should continue pipeline
            execution and expect no response from the client.

        Args:
            width: The number of cells in the buffer width.
            height: The number of cells in the buffer height.

        .. _PSHostRawUserInterface.BufferSize Property:
            https://docs.microsoft.com/en-us/dotnet/api/system.management.automation.host.pshostrawuserinterface.buffersize
        """
        self._connection.host_call(
            HostMethodIdentifier.SetBufferSize, [Size.ToPSObjectForRemoting(Size(width, height), for_host=True)]
        )

    def get_window_size(
        self,
    ) -> int:
        """GetWindowSize Request.

        Sends the
        :class:`psrpcore.types.HostMethodIdentifier.GetWindowSize` host
        call to get the size of the view window as measured by cells.

        This corresponds to the
        `PSHostRawUserInterface.WindowSize Property`_.

        Note:
            The server should wait for the host response before continuing the
            pipeline that created this request.

        Returns:
            int: The call id for the request.

        .. _PSHostRawUserInterface.WindowSize Property:
            https://docs.microsoft.com/en-us/dotnet/api/system.management.automation.host.pshostrawuserinterface.windowsize
        """
        return self._connection.host_call(HostMethodIdentifier.GetWindowSize)

    def set_window_size(
        self,
        width: int,
        height: int,
    ) -> None:
        """SetWindowSize Request.

        Sends the
        :class:`psrpcore.types.HostMethodIdentifier.SetWindowSize` host
        call to change the size of the view window size as measured by cells.

        This corresponds to the
        `PSHostRawUserInterface.WindowSize Property`_.

        Note:
            This is a void method and the server should continue pipeline
            execution and expect no response from the client.

        Args:
            width: The number of cells in the view window width.
            height: The number of cells in the view window height.

        .. _PSHostRawUserInterface.WindowSize Property:
            https://docs.microsoft.com/en-us/dotnet/api/system.management.automation.host.pshostrawuserinterface.windowsize
        """
        self._connection.host_call(
            HostMethodIdentifier.SetWindowSize, [Size.ToPSObjectForRemoting(Size(width, height), for_host=True)]
        )

    def get_window_title(
        self,
    ) -> int:
        """GetWindowTitle Request.

        Sends the
        :class:`psrpcore.types.HostMethodIdentifier.GetWindowTitle` host
        call to get the window title.

        This corresponds to the
        `PSHostRawUserInterface.WindowTitle Property`_.

        Note:
            The server should wait for the host response before continuing the
            pipeline that created this request.

        Returns:
            int: The call id for the request.

        .. _PSHostRawUserInterface.WindowTitle Property:
            https://docs.microsoft.com/en-us/dotnet/api/system.management.automation.host.pshostrawuserinterface.windowtitle
        """
        return self._connection.host_call(HostMethodIdentifier.GetWindowTitle)

    def set_window_title(
        self,
        title: str,
    ) -> None:
        """SetWindowTitle Request.

        Sends the
        :class:`psrpcore.types.HostMethodIdentifier.SetWindowTitle` host
        call to change the title of the host window

        This corresponds to the
        `PSHostRawUserInterface.WindowTitle Property`_.

        Note:
            This is a void method and the server should continue pipeline
            execution and expect no response from the client.

        Args:
            title: The title to set the host window to.

        .. _PSHostRawUserInterface.WindowTitle Property:
            https://docs.microsoft.com/en-us/dotnet/api/system.management.automation.host.pshostrawuserinterface.windowtitle
        """
        self._connection.host_call(HostMethodIdentifier.SetWindowTitle, [title])

    def get_max_window_size(
        self,
    ) -> int:
        """GetMaxWindowSize Request.

        Sends the
        :class:`psrpcore.types.HostMethodIdentifier.GetMaxWindowSize` host
        call to get largest possible window size that can fit in the current
        buffer.

        This corresponds to the
        `PSHostRawUserInterface.MaxWindowSize Property`_.

        Note:
            The server should wait for the host response before continuing the
            pipeline that created this request.

        Returns:
            int: The call id for the request.

        .. _PSHostRawUserInterface.MaxWindowSize Property:
            https://docs.microsoft.com/en-us/dotnet/api/system.management.automation.host.pshostrawuserinterface.maxwindowsize
        """
        return self._connection.host_call(HostMethodIdentifier.GetMaxWindowSize)

    def get_max_physical_window_size(
        self,
    ) -> int:
        """GetMaxPhysicalWindowSize Request.

        Sends the
        :class:`psrpcore.types.HostMethodIdentifier.GetMaxPhysicalWindowSize`
        host call to get largest possible window size that can be set.

        This corresponds to the
        `PSHostRawUserInterface.MaxPhysicalWindowSize Property`_.

        Note:
            The server should wait for the host response before continuing the
            pipeline that created this request.

        Returns:
            int: The call id for the request.

        .. _PSHostRawUserInterface.MaxPhysicalWindowSize Property:
            https://docs.microsoft.com/en-us/dotnet/api/system.management.automation.host.pshostrawuserinterface.maxphysicalwindowsize
        """
        return self._connection.host_call(HostMethodIdentifier.GetMaxPhysicalWindowSize)

    def get_key_available(
        self,
    ) -> int:
        """GetKeyAvailable Request.

        Sends the
        :class:`psrpcore.types.HostMethodIdentifier.GetKeyAvailable` host
        call to examine if a keystroke is waiting in the input buffer.

        This corresponds to the
        `PSHostRawUserInterface.KeyAvailable Property`_.

        Note:
            The server should wait for the host response before continuing the
            pipeline that created this request.

        Returns:
            int: The call id for the request.

        .. _PSHostRawUserInterface.KeyAvailable Property:
            https://docs.microsoft.com/en-us/dotnet/api/system.management.automation.host.pshostrawuserinterface.keyavailable
        """
        return self._connection.host_call(HostMethodIdentifier.GetKeyAvailable)

    def read_key(
        self,
        options: ReadKeyOptions = ReadKeyOptions.IncludeKeyDown,
    ) -> int:
        """ReadKey Request.

        Sends the :class:`psrpcore.types.HostMethodIdentifier.ReadKey` host
        call to request read a key stroke.

        This corresponds to the
        `PSHostRawUserInterface.ReadKey Method`_.

        Note:
            The server should wait for the host response before continuing the
            pipeline that created this request.

        Args:
            options: Further options to control the read key operation.

        Returns:
            int: The call id for the request.

        .. _PSHostRawUserInterface.ReadKey Method:
            https://docs.microsoft.com/en-us/dotnet/api/system.management.automation.host.pshostrawuserinterface.readkey
        """
        return self._connection.host_call(HostMethodIdentifier.ReadKey, [options.value])

    def flush_input_buffer(
        self,
    ) -> None:
        """FlushInputBuffer Request.

        Sends the
        :class:`psrpcore.types.HostMethodIdentifier.FlushInputBuffer` host
        call to reset the keyboard input buffer.

        This corresponds to the
        `PSHostRawUserInterface.FlushInputBuffer Method`_.

        Note:
            This is a void method and the server should continue pipeline
            execution and expect no response from the client.

        .. _PSHostRawUserInterface.FlushInputBuffer Method:
            https://docs.microsoft.com/en-us/dotnet/api/system.management.automation.host.pshostrawuserinterface.flushinputbuffer
        """
        self._connection.host_call(HostMethodIdentifier.FlushInputBuffer)

    def set_buffer_cells(
        self,
        left: int,
        top: int,
        right: int,
        bottom: int,
        character: typing.Union[int, str, PSChar],
        foreground: ConsoleColor = ConsoleColor.White,
        background: ConsoleColor = ConsoleColor.Black,
    ) -> None:
        """SetBufferContents Request.

        Sends the
        :class:`psrpcore.types.HostMethodIdentifier.SetBufferContents1` host
        call to set the contents of the host buffer.

        This corresponds to the
        `PSHostRawUserInterface.SetBufferContents Method`_. See
        :meth:`set_buffer_contents()` to set the buffer contents by individual
        cells.

        Note:
            This is a void method and the server should continue pipeline
            execution and expect no response from the client.

        Args:
            left: The left margin of the region to set the cell to.
            top: The top margin of the region to set the cell to.
            right: The right margin of the region to set the cell to.
            bottom: The bottom margin of the region to set the cell to.
            character: The character used to fill the cells in the region
                specified.
            foreground: The foreground (text) color to fill the cells in the
                region specified.
            background: The background color to fill the cells in the region
                specified.

        .. _PSHostRawUserInterface.SetBufferContents Method:
            https://docs.microsoft.com/en-us/dotnet/api/system.management.automation.host.pshostrawuserinterface.setbuffercontents
        """
        rectangle = Rectangle.ToPSObjectForRemoting(Rectangle(left, top, right, bottom), for_host=True)
        cell = BufferCell.ToPSObjectForRemoting(
            BufferCell(character, foreground, background, BufferCellType.Complete), for_host=True
        )
        self._connection.host_call(HostMethodIdentifier.SetBufferContents1, [rectangle, cell])

    def set_buffer_contents(
        self,
        x: int,
        y: int,
        contents: typing.List[typing.List[BufferCell]],
    ) -> None:
        """SetBufferContents Request.

        Sends the
        :class:`psrpcore.types.HostMethodIdentifier.SetBufferContents2` host
        call to set the contents of the host buffer.

        The ``contents`` value is a list of a list of
        :class:`psrpcore.types.BufferCell` where the first list dimension
        represents each row and the 2nd dimension is each column of that row.

        This corresponds to the
        `PSHostRawUserInterface.SetBufferContents Method`_. See
        :meth:`set_buffer_cells()` to set an individual cell across a region.

        Note:
            This is a void method and the server should continue pipeline
            execution and expect no response from the client.

        Args:
            x: The horizontal location of the upper left corner of the region
                to write the cells from.
            y: The vertical location of the upper left corner of the region to
                write the cells from.
            contents: A list of a list of cells that should be written.

        .. _PSHostRawUserInterface.SetBufferContents Method:
            https://docs.microsoft.com/en-us/dotnet/api/system.management.automation.host.pshostrawuserinterface.setbuffercontents
        """
        coordinates = Coordinates.ToPSObjectForRemoting(Coordinates(x, y), for_host=True)
        wrap_cell = functools.partial(BufferCell.ToPSObjectForRemoting, for_host=True)
        host_contents = _create_multi_dimensional_array(contents, wrap_cell)
        self._connection.host_call(HostMethodIdentifier.SetBufferContents2, [coordinates, host_contents])

    def get_buffer_contents(
        self,
        left: int,
        top: int,
        right: int,
        bottom: int,
    ) -> int:
        """GetBufferContents Request.

        Sends the
        :class:`psrpcore.types.HostMethodIdentifier.GetBufferContents` host
        call to request the contents of the screen region specified.

        This corresponds to the
        `PSHostRawUserInterface.GetBufferContents Method`_.

        Note:
            The server should wait for the host response before continuing the
            pipeline that created this request.

        Note:
            PowerShell does not implement this call for security reasons.
            Unless you trust the client host you should write an error record
            instead.

        Args:
            left: The left margin of the buffer region.
            top: The top margin of the buffer region.
            right: The right margin of the buffer region.
            bottom: The bottom margin of the buffer region.

        Returns:
            int: The call id for the request.

        .. _PSHostRawUserInterface.GetBufferContents Method:
            https://docs.microsoft.com/en-us/dotnet/api/system.management.automation.host.pshostrawuserinterface.getbuffercontents
        """
        return self._connection.host_call(
            HostMethodIdentifier.GetBufferContents,
            [
                Rectangle.ToPSObjectForRemoting(Rectangle(left, top, right, bottom), for_host=True),
            ],
        )

    def scroll_buffer_contents(
        self,
        source_left: int,
        source_top: int,
        source_right: int,
        source_bottom: int,
        x: int,
        y: int,
        clip_left: int,
        clip_top: int,
        clip_right: int,
        clip_bottom: int,
        character: typing.Union[int, str, PSChar],
        foreground: ConsoleColor = ConsoleColor.White,
        background: ConsoleColor = ConsoleColor.Black,
    ) -> None:
        """ScrollBufferContents Request.

        Sends the
        :class:`psrpcore.types.HostMethodIdentifier.ScrollBufferContents` host
        call to scoll a region of the screen buffer.

        This corresponds to the
        `PSHostRawUserInterface.ScrollBufferContents Method`_.

        Note:
            This is a void method and the server should continue pipeline
            execution and expect no response from the client.

        Args:
            source_left: The left margin of the screen to be scrolled.
            source_top: The top margin of the screen to be scrolled.
            source_right: The right margin of the screen to be scrolled.
            source_bottom: The bottom margin of the screen to be scrolled.
            x: The horizontal location of the upper left coordinate to receive
                the source region contents.
            y: The vertical location of the upper left coordinate to receive
                the source region contents.
            clip_left: The left margin of the clipped region.
            clip_top: The top margin of the clipped region.
            clip_right: The right margin of the clipped region.
            clip_bottom: The bottom margin of the clipped region.
            character: The character used to fill the cells intersecting the
                source and clip region.
            foreground: The foreground (text) color to fill the cells
                intersecting the source and clip region.
            background: The background color to fill the cells intersecting
                the source and clip region.

        .. _PSHostRawUserInterface.ScrollBufferContents Method:
            https://docs.microsoft.com/en-us/dotnet/api/system.management.automation.host.pshostrawuserinterface.scrollbuffercontents
        """
        source = Rectangle.ToPSObjectForRemoting(
            Rectangle(source_left, source_top, source_right, source_bottom), for_host=True
        )
        dest = Coordinates.ToPSObjectForRemoting(Coordinates(x, y), for_host=True)
        clip = Rectangle.ToPSObjectForRemoting(Rectangle(clip_left, clip_top, clip_right, clip_bottom), for_host=True)
        fill = BufferCell.ToPSObjectForRemoting(
            BufferCell(character, foreground, background, BufferCellType.Complete), for_host=True
        )

        self._connection.host_call(HostMethodIdentifier.ScrollBufferContents, [source, dest, clip, fill])



# -*- coding: utf-8 -*-
# Copyright: (c) 2021, Jordan Borean (@jborean93) <jborean93@gmail.com>
# MIT License (see LICENSE or https://opensource.org/licenses/MIT)

"""PSRP/.NET Complex Types.

The PSRP/.NET Complex Type class definitions. A complex type is pretty much
anything that isn't a primitive type as known to the protocol. Most of the
types defined here are defined in `MS-PSRP 2.2.3 Other Object Types`_ but some
are also just other .NET objects that are used in the PSRP protocol. Some types
are a PSRP specific representation of an actual .NET type but with a few minor
differences. These types are prefixed with `PSRP` to differentiate between the
PSRP specific ones and the actual .NET types.

.. _MS-PSRP 2.2.3 Other Object Types:
    https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-psrp/e41c4a38-a821-424b-bc1c-89f8478c39ae
"""

import collections
import ntpath
import posixpath
import typing

from psrpcore.types._base import (
    PSAliasProperty,
    PSNoteProperty,
    PSObject,
    PSScriptProperty,
    PSType,
    add_note_property,
)
from psrpcore.types._collection import PSDict, PSList
from psrpcore.types._enum import PSEnumBase, PSFlagBase
from psrpcore.types._primitive import (
    PSBool,
    PSDateTime,
    PSInt,
    PSInt64,
    PSSecureString,
    PSString,
    PSUInt,
)

# We are just using a named tuple for now, this should be a class in the future
# if CommandInfo is ever implemented.
RemoteCommandInfo = collections.namedtuple("RemoteCommandInfo", ["CommandType", "Name", "Definition", "Visibility"])


@PSType(["System.Management.Automation.PSCustomObject"])
class PSCustomObject(PSObject):
    """PSCustomObject

    This is a PSCustomObject that can be created with an arbitrary amount of
    extended properties. It acts like a generic property bag and is designed to
    replicate the PSCustomObject syntax in PowerShell:

    .. code-block:: powershell

        $obj = [PSCustomObject]@{
            Property = 'value'
        }

    Examples:
        >>> obj = PSCustomObject(Property='value')
        >>> print(obj.Property)
        abc

    Note:
        Using `PSTypeName` as a parameter will define a custom PS type name for
        the instance. It will not add a property with the name `PSTypeName`.
    """

    def __init__(self, **kwargs: typing.Any) -> None:
        for prop_name, prop_value in kwargs.items():
            # Special use case with [PSCustomObject]@{PSTypeName = 'TypeName'} in PowerShell where the value is
            # added to the top of the objects type names.
            if prop_name == "PSTypeName":
                self.PSObject.type_names.insert(0, prop_value)

            else:
                self.PSObject.extended_properties.append(PSNoteProperty(prop_name, value=prop_value))


@PSType(["System.ConsoleColor"])
class ConsoleColor(PSEnumBase):
    """System.ConsoleColor enum.

    Specifies constants that define foreground and background colors for the
    console. This is also documented under `[MS-PSRP] 2.2.3.3 Color`_ but in
    the :class:`HostInfo` default data format.

    Note:
        This is an auto-generated Python class for the `System.ConsoleColor`_
        .NET class.

    .. _System.ConsoleColor:
        https://docs.microsoft.com/en-us/dotnet/api/system.consolecolor?view=net-5.0

    .. _[MS-PSRP] 2.2.3.3 Color:
        https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-psrp/d7edefec-41b1-465d-bc07-2a8ec9d727a1
    """

    Black = 0
    DarkBlue = 1
    DarkGreen = 2
    DarkCyan = 3
    DarkRed = 4
    DarkMagenta = 5
    DarkYellow = 6
    Gray = 7
    DarkGray = 8
    Blue = 9
    Green = 10
    Cyan = 11
    Red = 12
    Magenta = 13
    Yellow = 14
    White = 15


@PSType(["System.Management.Automation.ProgressRecordType"])
class ProgressRecordType(PSEnumBase):
    """System.Management.Automation.ProgressRecordType enum.

    Defines two types of progress record that refer to the beginning
    (or middle) and end of an operation.

    Note:
        This is an auto-generated Python class for the
        `System.Management.Automation.ProgressRecordType`_ .NET class.

    .. _System.Management.Automation.ProgressRecordType:
        https://docs.microsoft.com/en-us/dotnet/api/system.management.automation.progressrecordtype
    """

    Processing = 0  #: Operation just started or is not yet complete.
    Completed = 1  #: Operation is complete.


@PSType(["System.Management.Automation.PSCredentialTypes"])
class PSCredentialTypes(PSFlagBase):
    """System.Management.Automation.PSCredentialTypes enum flags.

    Defines the valid types of credentials. Used by
    :class:`HostMethodIdentifier.PromptForCredential2` calls.

    Note:
        This is an auto-generated Python class for the
        `System.Management.Automation.PSCredentialTypes`_ .NET class.

    .. _System.Management.Automation.PSCredentialTypes:
        https://docs.microsoft.com/en-us/dotnet/api/system.management.automation.pscredentialtypes
    """

    Generic = 1  #: Generic credentials.
    Domain = 2  #: Credentials valid for a domain.
    Default = Generic | Domain  #: Default credentials (Generic | Domain).


@PSType(["System.Management.Automation.PSCredentialUIOptions"])
class PSCredentialUIOptions(PSFlagBase):
    """System.Management.Automation.PSCredentialUIOptions enum flags.

    Defines the options available when prompting for credentials. Used by
    :class:`HostMethodIdentifier.PromptForCredential2` calls.

    Note:
        This is an auto-generated Python class for the
        `System.Management.Automation.PSCredentialUIOptions`_ .NET class.

    .. _System.Management.Automation.PSCredentialUIOptions:
        https://docs.microsoft.com/en-us/dotnet/api/system.management.automation.pscredentialuioptions
    """

    none = 0  #: Performs no validation.
    ValidateUserNameSyntax = 1  #: Validates the username, but not its existence or correctness.
    AlwaysPrompt = 2  #: Always prompt, even if a persisted credential was available.
    ReadOnlyUsername = 3  #: Username is read-only, and the user may not modify it.
    Default = ValidateUserNameSyntax  #: Validates the username, but not its existence or correctness.


@PSType(["System.Management.Automation.SessionStateEntryVisibility"])
class SessionStateEntryVisibility(PSEnumBase):
    """System.Management.Automation.SessionStateEntryVisibility enum.

    Defines the visibility of execution environment elements.

    Note:
        This is an auto-generated Python class for the
        `System.Management.Automation.SessionStateEntryVisibility`_ .NET class.

    .. _System.Management.Automation.SessionStateEntryVisibility:
        https://docs.microsoft.com/en-us/dotnet/api/system.management.automation.sessionstateentryvisibility
    """

    Public = 0  #: Entries are visible to requests from outside the runspace.
    Private = 1  #: Entries are not visible to requests from outside the runspace.


@PSType(["System.Management.Automation.Runspaces.RunspacePoolState"])
class RunspacePoolState(PSEnumBase):
    """RunspacePoolState enum.

    Defines the current state of the Runspace Pool. It is documented in PSRP
    under `[MS-PSRP] 2.2.3.4 RunspacePoolState`_ and while it shares the same
    name as the .NET type
    `System.Management.Automation.Runspaces.RunspacePoolState`_ it has a few
    extra names and some values that do not match. The .NET enum is used here
    as that's what PowerShell uses internally.

    .. _[MS-PSRP] 2.2.3.4 RunspacePoolState:
        https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-psrp/b05495bc-a9b2-4794-9f43-4bf1f3633900

    .. _System.Management.Automation.Runspaces.RunspacePoolState:
        https://docs.microsoft.com/en-us/dotnet/api/system.management.automation.runspaces.runspacepoolstate
    """

    BeforeOpen = 0  #: Beginning state upon creation.
    Opening = 1  #: A RunspacePool is being created.
    Opened = 2  #: The RunspacePool is created and valid.
    Closed = 3  #: The RunspacePool is closed.
    Closing = 4  #: The RunspacePool is being closed.
    Broken = 5  #: The RunspacePool has been disconnected abnormally.
    Disconnecting = 6  #: The RunspacePool is being disconnected.
    Disconnected = 7  #: The RunspacePool has been disconnected.
    Connecting = 8  #: The RunspacePool is being connected.


@PSType(["System.Management.Automation.PSInvocationState"])
class PSInvocationState(PSEnumBase):
    """PSInvocationState enum.

    Defines the current state of the Pipeline. It is documented in PSRP under
    `[MS-PSRP] 2.2.3.5 PSInvocationState`_.

    Note:
        This is an auto-generated Python class for the
        `System.Management.Automation.PSInvocationState`_ .NET class.

    .. _[MS-PSRP] 2.2.3.5 PSInvocationState:
        https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-psrp/acaa253a-29be-45fd-911c-6715515a28b9

    .. _System.Management.Automation.PSInvocationState:
        https://docs.microsoft.com/en-us/dotnet/api/system.management.automation.psinvocationstate
    """

    NotStarted = 0  #: Pipeline has not been started
    Running = 1  #: Pipeline is executing.
    Stopping = 2  #: Pipeline is stopping execution.
    Stopped = 3  #: Pipeline is completed due to a stop request.
    Completed = 4  #: Pipeline has completed executing a command.
    Failed = 5  #: Pipeline completed abnormally due to an error.
    Disconnected = 6  #: Pipeline is in disconnected state.


@PSType(["System.Management.Automation.Runspaces.PSThreadOptions"])
class PSThreadOptions(PSEnumBase):
    """System.Management.Automation.Runspaces.PSThreadOptions enum.

    Control whether a new thread is created when a command is executed within a
    Runspace. It is documented in PSRP under
    `[MS-PSRP] 2.2.3.6 PSThreadOptions`_.

    Note:
        This is an auto-generated Python class for the
        `System.Management.Automation.Runspaces.PSThreadOptions`_ .NET class.

    .. _[MS-PSRP] 2.2.3.6 PSThreadOptions:
        https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-psrp/bfc63adb-d6f1-4ccc-9bd8-73de6cc78dda

    .. _System.Management.Automation.Runspaces.PSThreadOptions:
        https://docs.microsoft.com/en-us/dotnet/api/system.management.automation.runspaces.psthreadoptions
    """

    Default = 0  #: Use the server thread option settings.
    UseNewThread = 1  #: Creates a new thread for each invocation.
    ReuseThread = 2
    """ Creates a new thread for the first invocation and then re-uses that thread in subsequent invocations. """
    UseCurrentThread = 3  #: Doesn't create a new thread; the execution occurs on the thread that called Invoke.


@PSType(["System.Threading.ApartmentState"])
class ApartmentState(PSEnumBase):
    """System.Management.Automation.Runspaces.ApartmentState enum.

    Specifies the apartment state of a Thread. It is documented in PSRP under
    `[MS-PSRP] 2.2.3.7 ApartmentState`_.

    Note:
        This is an auto-generated Python class for the
        `System.Threading.ApartmentState`_ .NET class.

    .. _[MS-PSRP] 2.2.3.7 ApartmentState:
        https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-psrp/6845133d-7503-450d-a74e-388cdd3b2386

    .. _System.Threading.ApartmentState:
        https://docs.microsoft.com/en-us/dotnet/api/system.threading.apartmentstate?view=net-5.0
    """

    STA = 0  #: The thread will create and enter a multi-threaded apartment.
    MTA = 1  #: The thread will create and enter a single-threaded apartment.
    Unknown = 2  #: The ApartmentState property has not been set.


@PSType(["System.Management.Automation.RemoteStreamOptions"])
class RemoteStreamOptions(PSFlagBase):
    """System.Management.Automation.RemoteStreamOptions enum flags.

    Control whether InvocationInfo is added to items in the Error, Warning,
    Verbose and Debug streams during remote calls. It is documented in PSRP
    under `[MS-PSRP] 2.2.3.8 RemoteStreamOptions`_.

    Note:
        This is an auto-generated Python class for the
        `System.Management.Automation.RemoteStreamOptions`_ .NET class.

    .. _[MS-PSRP] 2.2.3.8 RemoteStreamOptions:
        https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-psrp/4941e59c-ce01-4549-8eb5-372b8eb6dd12

    .. _System.Management.Automation.RemoteStreamOptions:
        https://docs.microsoft.com/en-us/dotnet/api/system.management.automation.remotestreamoptions
    """

    none = 0  #: InvocationInfo is not added to any stream record.
    AddInvocationInfoToErrorRecord = 1  #: InvocationInfo is added to any :class:`ErrorRecord`.
    AddInvocationInfoToWarningRecord = 2  #: InvocationInfo is added to any `Warning` :class:`InformationalRecord`.
    AddInvocationInfoToDebugRecord = 4  #: InvocationInfo is added to any `Debug` :class:`InformationalRecord`.
    AddInvocationInfoToVerboseRecord = 8  #: InvocationInfo is added to any `Verbose` :class:`InformationalRecord`.
    AddInvocationInfo = 15  #: InvocationInfo is added to all stream records.


@PSType(["System.Management.Automation.ErrorCategory"])
class ErrorCategory(PSEnumBase):
    """System.Management.Automation.ErrorCategory enum.

    Errors reported by PowerShell will be in one of these categories. It is
    documented in PSRP under `[MS-PSRP] 2.2.3.9 ErrorCategory`_.

    Note:
        This is an auto-generated Python class for the
        `System.Management.Automation.ErrorCategory`_ .NET class.

    .. _[MS-PSRP] 2.2.3.9 ErrorCategory:
        https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-psrp/ae7d6061-15c8-4184-a05e-1033dbb7228b

    .. _System.Management.Automation.ErrorCategory:
        https://docs.microsoft.com/en-us/dotnet/api/system.management.automation.errorcategory
    """

    NotSpecified = 0  #: No error category is specified, or the error category is invalid.
    OpenError = 1
    CloseError = 2
    DeviceError = 3
    DeadlockDetected = 4
    InvalidArgument = 5
    InvalidData = 6
    InvalidOperation = 7
    InvalidResult = 8
    InvalidType = 9
    MetadataError = 10
    NotImplemented = 11
    NotInstalled = 12
    ObjectNotFound = 13  #: Object can not be found (file, directory, computer, system resource, etc.).
    OperationStopped = 14
    OperationTimeout = 15
    SyntaxError = 16
    ParserError = 17
    PermissionDenied = 18  #: Operation not permitted.
    ResourceBusy = 19
    ResourceExists = 20
    ResourceUnavailable = 21
    ReadError = 22
    WriteError = 23
    FromStdErr = 24  #: A non-PowerShell command reported an error to its STDERR pipe.
    SecurityError = 25  #: Used for security exceptions.
    ProtocolError = 26  #: The contract of a protocol is not being followed.
    ConnectionError = 27  #: The operation depends on a network connection that cannot be established or maintained.
    AuthenticationError = 28  #: Could not authenticate the user to the service.
    LimitsExceeded = 29  #: Internal limits prevent the operation from being executed.
    QuotaExceeded = 30  #: Controls on the use of traffic or resources prevent the operation from being executed.
    NotEnabled = 31  #: The operation attempted to use functionality that is currently disabled.


@PSType(["System.Management.Automation.CommandTypes"])
class CommandTypes(PSFlagBase):
    """System.Management.Automation.CommandTypes enum flags.

    Defines the types of commands that PowerShell can execute. It is documented
    in PSRP under `[MS-PSRP] 2.2.3.19 CommandType`_.

    Note:
        This is an auto-generated Python class for the
        `System.Management.Automation.CommandTypes`_ .NET class.

    .. _[MS-PSRP] 2.2.3.19 CommandType:
        https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-psrp/a038c5c9-a220-4064-aa78-ed9cf5a2893c

    .. _System.Management.Automation.CommandTypes:
        https://docs.microsoft.com/en-us/dotnet/api/system.management.automation.commandtypes
    """

    Alias = 1  #: Aliases create a name that refers to other command types.
    Function = 2  #: Script functions that are defined by a script block.
    Filter = 4  #: Script filters that are defined by a script block.
    Cmdlet = 8  #: A cmdlet.
    ExternalScript = 16  #: A PowerShell script (`*.ps1` file).
    Application = 32  #: Any existing application (can be console or GUI).
    Script = 64  #: A script that is built into the runspace configuration.
    Configuration = 256  #: A configuration.
    All = 383  #: All possible command types.


@PSType(["System.Management.Automation.CommandOrigin"])
class CommandOrigin(PSEnumBase):
    """System.Management.Automation.CommandOrigin enum.

    Defines the dispatch origin of a command. It is documented in PSRP under
    `[MS-PSRP] 2.2.3.30 CommandOrigin`_.

    Note:
        This is an auto-generated Python class for the
        `System.Management.Automation.CommandOrigin`_ .NET class.

    .. _[MS-PSRP] 2.2.3.30 CommandOrigin:
        https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-psrp/6c35a5de-d063-4097-ace5-002a0c5e452d

    .. _System.Management.Automation.CommandOrigin:
        https://docs.microsoft.com/en-us/dotnet/api/system.management.automation.commandorigin
    """

    Runspace = 0  #: The command was submited via a runspace.
    Internal = 1  #: The command was dispatched by the PowerShell engine.


@PSType(["System.Management.Automation.Runspaces.PipelineResultTypes"])
class PipelineResultTypes(PSFlagBase):
    """System.Management.Automation.Runspaces.PipelineResultTypes enum flags.

    Defines the types of streams coming out of a pipeline. It is documented in
    PSRP under `[MS-PSRP] 2.2.3.31 PipelineResultTypes`_. .NET and MS-PSRP have
    separate values but .NET is used as it is the correct source.
    Technically the values are not designed as flags but there are some older
    APIs that combine Output | Error together.

    Note:
        This is an auto-generated Python class for the
        `System.Management.Automation.Runspaces.PipelineResultTypes`_ .NET
        class.

    .. _[MS-PSRP] 2.2.3.31 PipelineResultTypes:
        https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-psrp/efdce0ba-531e-4904-9cab-b65c476c649a

    .. _System.Management.Automation.Runspaces.PipelineResultTypes:
        https://docs.microsoft.com/en-us/dotnet/api/system.management.automation.runspaces.pipelineresulttypes
    """

    none = 0  #: Default streaming behaviour.
    Output = 1  #: Output stream.
    Error = 2  #: Error stream.
    Warning = 3  #: Warning stream.
    Verbose = 4  #: Verbose stream.
    Debug = 5  #: Debug stream.
    Information = 6  #: Information stream.
    All = 7  #: All streams.
    Null = 8  #: Redirect to nothing.


@PSType(
    type_names=[
        "System.Management.Automation.Language.ScriptPosition",
    ],
    adapted_properties=[
        PSNoteProperty("File", ps_type=PSString),
        PSNoteProperty("LineNumber", ps_type=PSInt),
        PSNoteProperty("ColumnNumber", ps_type=PSInt),
        PSNoteProperty("Line", ps_type=PSString),
        PSScriptProperty("Offset", lambda o: 0, ps_type=PSInt),  # Not used in pwsh.
    ],
)
class ScriptPosition(PSObject):
    """ScriptPosition.

    Represents a single point in a script. This script may come from a file or
    interactive input. This is the actual .NET type
    `System.Management.Automation.Language.ScriptPosition`_.

    Args:
        File: The name of the file, or if the script did not come from a file,
            then null.
        LineNumber: The line number of the position, with the value 1 being the
            first line.
        ColumnNumber: The column number of the position, with the value 1 being
            the first column.
        Line: The complete text of the line that this position is included on.

    Attributes:
        Offset (:class:`PSInt`): Not used in pwsh.

    .. _System.Management.Automation.Language.ScriptPosition:
        https://docs.microsoft.com/en-us/dotnet/api/system.management.automation.language.scriptposition
    """


@PSType(
    type_names=[
        "System.Management.Automation.Language.ScriptExtent",
    ],
    adapted_properties=[
        PSScriptProperty("File", lambda o: o._start.File, ps_type=PSString),
        PSScriptProperty("StartScriptPosition", lambda o: o._start, ps_type=PSString),
        PSScriptProperty("EndScriptPosition", lambda o: o._end, ps_type=PSString),
        PSScriptProperty("StartLineNumber", lambda o: o._start.LineNumber, ps_type=PSInt),
        PSScriptProperty("StartColumnNumber", lambda o: o._start.ColumnNumber, ps_type=PSInt),
        PSScriptProperty("EndLineNumber", lambda o: o._end.LineNumber, ps_type=PSInt),
        PSScriptProperty("EndColumnNumber", lambda o: o._end.ColumnNumber, ps_type=PSInt),
        PSScriptProperty("StartOffset", lambda o: o._start.Offset, ps_type=PSInt),
        PSScriptProperty("EndOffset", lambda o: o._end.Offset, ps_type=PSInt),
        PSAliasProperty("Text", "_text", ps_type=PSString),
    ],
)
class ScriptExtent(PSObject):
    """ScriptExtent.

    A script extend used to customize the display of error location
    information. This is the actual .NET type
    `System.Management.Automation.Language.ScriptExtent`_.

    Args:
        StartPosition: The position beginning the extent.
        EndPosition: The position ending the extent.

    .. _System.Management.Automation.Language.ScriptExtent:
        https://docs.microsoft.com/en-us/dotnet/api/system.management.automation.language.scriptextent
    """

    def __init__(
        self,
        StartPosition: ScriptPosition,
        EndPosition: ScriptPosition,
    ):
        super().__init__()
        self._start = StartPosition
        self._end = EndPosition

    @property
    def _text(self) -> PSString:
        if self.EndColumnNumber < 1:
            val = ""

        elif self.StartLineNumber == self.EndLineNumber:
            val = self._start.Line[max(self.StartColumnNumber - 1, 0) : self.EndColumnNumber]

        else:
            start = self._start.Line[self.StartColumnNumber :]
            end = self._end.Line[self.EndColumnNumber :]
            val = f"{start}...{end}"

        return PSString(val)


def _script_extent_from_ps_object(
    obj: PSObject,
) -> ScriptExtent:
    """Used by ErrorRecord and InformationalRecord to deserialize the invocation info details."""
    file = obj.ScriptExtent_File
    start_line = obj.ScriptExtent_StartLineNumber
    end_line = obj.ScriptExtent_EndLineNumber
    start_column = obj.ScriptExtent_StartColumnNumber
    end_column = obj.ScriptExtent_EndColumnNumber

    return ScriptExtent(
        StartPosition=ScriptPosition(file, start_line, start_column, ""),
        EndPosition=ScriptPosition(file, end_line, end_column, ""),
    )


def _script_extent_to_ps_object(
    script_extent: ScriptExtent,
    obj: PSObject,
) -> None:
    """Used by InvocationInfo to serialize the script extent details."""
    add_note_property(obj, "ScriptExtent_File", script_extent.File)
    add_note_property(obj, "ScriptExtent_StartLineNumber", script_extent.StartLineNumber)
    add_note_property(obj, "ScriptExtent_StartColumnNumber", script_extent.StartColumnNumber)
    add_note_property(obj, "ScriptExtent_EndLineNumber", script_extent.EndLineNumber)
    add_note_property(obj, "ScriptExtent_EndColumnNumber", script_extent.EndColumnNumber)


@PSType(
    type_names=[
        "System.Management.Automation.InvocationInfo",
    ],
    adapted_properties=[
        PSNoteProperty("BoundParameters", ps_type=PSDict),
        PSNoteProperty("CommandOrigin", ps_type=CommandOrigin),
        PSNoteProperty("DisplayScriptPosition", ps_type=ScriptExtent),
        PSNoteProperty("ExpectingInput", ps_type=PSBool),
        PSNoteProperty("HistoryId", ps_type=PSInt64),
        PSNoteProperty("InvocationName", ps_type=PSString),
        PSNoteProperty("Line", ps_type=PSString),
        PSNoteProperty("MyCommand"),  # CommandInfo,
        PSNoteProperty("OffsetInLine", ps_type=PSInt),
        PSNoteProperty("PipelineLength", ps_type=PSInt),
        PSNoteProperty("PipelinePosition", ps_type=PSInt),
        PSNoteProperty("PositionMessage", ps_type=PSString),
        PSNoteProperty("PSCommandPath", ps_type=PSString),
        PSNoteProperty("PSScriptRoot", ps_type=PSString),
        PSNoteProperty("ScriptLineNumber", ps_type=PSInt),
        PSNoteProperty("ScriptName", ps_type=PSString),
        PSNoteProperty("UnboundArguments", ps_type=PSList),
    ],
)
class InvocationInfo(PSObject):
    """InvocationInfo.

    Describes how and where this command was invoked. This is the actual .NET
    type `System.Management.Automation.InvocationInfo`_.

    Args:
        BoundParameters: Dictionary of parameters that were bound for this
            script or command.
        CommandOrigin: Command was being invoked inside the runspace or if it
            was an external request.
        DisplayScriptPosition: The position for the invocation or error.
        ExpectingInput: The command is expecting input.
        HistoryId: History that represents the command, if unavailable this
            will be ``-1``.
        InvocationName: Command name used to invoke this script. If invoked
            through an alias then this would be the alias name.
        Line: The text of the line that contained this cmdlet invocation.
        MyCommand: Basic information about the command.
        OffsetInLine: Command's character offset in that line. If the command
            was executed directly through the host interfaces, this will be -1.
        PipelineLength: How many elements are in the containing pipeline.
        PipelinePosition: Which element this command was in the containing
            pipeline.
        PositionMessage: Formatted message indicating where the cmdlet appeared
            in the line.
        PSCommandPath: The full path to the command from where it was being
            invoked.
        PSScriptRoot: The directory from where the command was being invoked.
        ScriptLineNumber: The line number in the executing script that contains
            this cmdlet.
        ScriptName: The name of the script containing the cmdlet.
        UnboundArguments: The list of arguments that were not bound to any
            parameter.

    .. _System.Management.Automation.InvocationInfo:
        https://docs.microsoft.com/en-us/dotnet/api/system.management.automation.invocationinfo
    """


def _invocation_info_from_ps_object(
    obj: PSObject,
) -> InvocationInfo:
    """Used by ErrorRecord and InformationalRecord to deserialize the invocation info details."""
    line = getattr(obj, "InvocationInfo_Line", None) or ""
    ps_command_path = ""
    ps_script_root = ""

    if getattr(obj, "SerializeExtent") or False:
        display_script_position = _script_extent_from_ps_object(obj)

    else:
        script_name = obj.InvocationInfo_ScriptName or ""
        script_line_number = obj.InvocationInfo_ScriptLineNumber
        offset_in_line = obj.InvocationInfo_OffsetInLine
        start_position = ScriptPosition(script_name, script_line_number, offset_in_line, line)

        end_position = start_position
        if line:
            end_position = ScriptPosition(script_name, script_line_number, len(line) + 1, line)

        display_script_position = ScriptExtent(start_position, end_position)

        ps_command_path = display_script_position.File
        if posixpath.sep in ps_command_path:  # pragma: no cover
            ps_script_root = posixpath.dirname(ps_command_path)
        else:  # pragma: no cover
            ps_script_root = ntpath.dirname(ps_command_path)

    my_command = None
    command_type = getattr(obj, "CommandInfo_CommandType", None)
    if command_type is not None:
        my_command = RemoteCommandInfo(
            CommandType=CommandTypes(command_type),
            Name=obj.CommandInfo_Name,
            Definition=obj.CommandInfo_Definition,
            Visibility=SessionStateEntryVisibility(obj.CommandInfo_Visibility),
        )

    return InvocationInfo(
        BoundParameters=getattr(obj, "InvocationInfo_BoundParameters", None) or PSDict(),
        CommandOrigin=CommandOrigin(obj.InvocationInfo_CommandOrigin),
        DisplayScriptPosition=display_script_position,
        ExpectingInput=obj.InvocationInfo_ExpectingInput,
        HistoryId=obj.InvocationInfo_HistoryId,
        InvocationName=obj.InvocationInfo_InvocationName,
        Line=line,
        MyCommand=my_command,
        OffsetInLine=obj.InvocationInfo_OffsetInLine,
        PipelineLength=obj.InvocationInfo_PipelineLength,
        PipelinePosition=obj.InvocationInfo_PipelinePosition,
        PSCommandPath=ps_command_path,
        PSScriptRoot=ps_script_root,
        ScriptLineNumber=obj.InvocationInfo_ScriptLineNumber,
        ScriptName=obj.InvocationInfo_ScriptName,
        UnboundArguments=getattr(obj, "InvocationInfo_UnboundArguments", None) or PSList(),
    )


def _invocation_info_to_ps_object(
    invocation_info: InvocationInfo,
    obj: PSObject,
) -> None:
    """Used by ErrorRecord and InformationalRecord to serialize the invocation info details."""
    add_note_property(obj, "InvocationInfo_BoundParameters", invocation_info.BoundParameters)
    add_note_property(obj, "InvocationInfo_CommandOrigin", invocation_info.CommandOrigin)
    add_note_property(obj, "InvocationInfo_ExpectingInput", invocation_info.ExpectingInput)
    add_note_property(obj, "InvocationInfo_InvocationName", invocation_info.InvocationName)
    add_note_property(obj, "InvocationInfo_Line", invocation_info.Line)
    add_note_property(obj, "InvocationInfo_OffsetInLine", invocation_info.OffsetInLine)
    add_note_property(obj, "InvocationInfo_HistoryId", invocation_info.HistoryId)
    add_note_property(obj, "InvocationInfo_PipelineIterationInfo", [])  # List of PSInt32
    add_note_property(obj, "InvocationInfo_PipelineLength", invocation_info.PipelineLength)
    add_note_property(obj, "InvocationInfo_PipelinePosition", invocation_info.PipelinePosition)
    add_note_property(obj, "InvocationInfo_PSScriptRoot", invocation_info.PSScriptRoot)
    add_note_property(obj, "InvocationInfo_PSCommandPath", invocation_info.PSCommandPath)
    add_note_property(obj, "InvocationInfo_PositionMessage", invocation_info.PositionMessage)
    add_note_property(obj, "InvocationInfo_ScriptLineNumber", invocation_info.ScriptLineNumber)
    add_note_property(obj, "InvocationInfo_ScriptName", invocation_info.ScriptName)
    add_note_property(obj, "InvocationInfo_UnboundArguments", invocation_info.UnboundArguments)

    if invocation_info.DisplayScriptPosition:
        _script_extent_to_ps_object(invocation_info.DisplayScriptPosition, obj)
        add_note_property(obj, "SerializeExtent", True)

    else:
        add_note_property(obj, "SerializeExtent", False)

    if invocation_info.MyCommand:
        # This should be a CommandInfo type but we are being pretty lax with the property checks.
        my_command = invocation_info.MyCommand
        add_note_property(obj, "CommandInfo_CommandType", getattr(my_command, "CommandType", CommandTypes.Application))
        add_note_property(obj, "CommandInfo_Definition", getattr(my_command, "Definition", ""))
        add_note_property(obj, "CommandInfo_Name", getattr(my_command, "Name", ""))
        add_note_property(
            obj, "CommandInfo_Visibility", getattr(my_command, "Visibility", SessionStateEntryVisibility.Public)
        )


@PSType(
    type_names=[
        "System.Management.Automation.ErrorCategoryInfo",
    ],
    adapted_properties=[
        # Technically a string in .NET but it's easier for the end user to be an enum.
        PSNoteProperty("Category", value=ErrorCategory.NotSpecified, ps_type=ErrorCategory),
        PSNoteProperty("Activity", ps_type=PSString),
        PSNoteProperty("Reason", ps_type=PSString),
        PSNoteProperty("TargetName", ps_type=PSString),
        PSNoteProperty("TargetType", ps_type=PSString),
    ],
)
class ErrorCategoryInfo(PSObject):
    """ErrorCategoryInfo.

    Contains auxiliary information about an ErrorRecord. This is the actual
    .NET type `System.Management.Automation.ErrorCategoryInfo`_.

    Args:
        Category: The error category.
        Activity: Description of the operation which encountered the error.
        Reason: Description of the error.
        TargetName: Description of the target object.
        TargetType: Description of the type of the target object.

    .. _System.Management.Automation.ErrorCategoryInfo:
        https://docs.microsoft.com/en-us/dotnet/api/system.management.automation.errorcategoryinfo
    """

    def __str__(self) -> str:
        return (
            f'{self.Category.name!s} ({self.TargetName or ""}:{self.TargetType or ""}) '
            f'[{self.Activity or ""}], {self.Reason or ""}'
        )


@PSType(
    type_names=[
        "System.Management.Automation.ErrorDetails",
    ],
    adapted_properties=[
        PSNoteProperty("Message", ps_type=PSString),
        PSNoteProperty("RecommendedAction", ps_type=PSString),
    ],
)
class ErrorDetails(PSObject):
    """ErrorDetails.

    ErrorDetails represents additional details about an :class:`ErrorRecord`,
    starting with a replacement Message. Clients can use ErrorDetails when they
    want to display a more specific message than the one contained in a
    particular Exception, without having to create a new Exception or define a
    new Exception class. This is the actual
    .NET type `System.Management.Automation.ErrorDetails`_.

    Args:
        Message: Message with replaces Message in Exception.
        RecommendedAction: Describes the recommended action in the event this
            error occurs. This can be empty if not applicable.

    .. _System.Management.Automation.ErrorDetails:
        https://docs.microsoft.com/en-us/dotnet/api/system.management.automation.errordetails
    """


@PSType(
    type_names=[
        "System.Management.Automation.ErrorRecord",
    ],
    adapted_properties=[
        PSNoteProperty("Exception", mandatory=True),
        PSNoteProperty("CategoryInfo", mandatory=True, ps_type=ErrorCategoryInfo),
        PSNoteProperty("TargetObject"),
        PSNoteProperty("FullyQualifiedErrorId", ps_type=PSString),
        PSNoteProperty("InvocationInfo", ps_type=InvocationInfo),
        PSNoteProperty("ErrorDetails", ps_type=ErrorDetails),
        PSNoteProperty("PipelineIterationInfo", ps_type=PSList),
        PSNoteProperty("ScriptStackTrace", ps_type=PSString),
    ],
)
class ErrorRecord(PSObject):
    """ErrorRecord.

    The data type that represents information about an error. It is documented
    in PSRP under `[MS-PSRP] 2.2.3.15 ErrorRecord`_. The invocation specific
    properties are documented under `[MS-PSRP] 2.2.3.15.1 InvocationInfo`_.
    This is the actual .NET type `System.Management.Automation.ErrorRecord`_

    Args:
        Exception: An exception describing the error.
        TargetObject: The object against which the error occurred.
        FullyQualifiedErrorId: String which uniquely identifies this error
            condition.
        CategoryInfo: Information regarding the ErrorCategory associated with
            this error.
        ErrorDetails: Additional information about the error.
        InvocationInfo: Identifies the cmdlet, script, or other command which
            caused the error.
        PipelineIterationInfo: The status of the pipeline when this record was
            created. Each entry represents the number of inputs the command[i]
            in the statement has processed when the record was created.
        ScriptStackTrace: The object against which the error occurred.

    .. _[MS-PSRP] 2.2.3.15 ErrorRecord:
        https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-psrp/0fe855a7-d13c-44e2-aa88-291e2054ae3a

    .. _[MS-PSRP] 2.2.3.15.1 InvocationInfo:
        https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-psrp/000363b7-e2f9-4a34-94f5-d540a15aee7b

    .. _System.Management.Automation.ErrorRecord:
        https://docs.microsoft.com/en-us/dotnet/api/system.management.automation.errorrecord
    """

    def __init__(
        self,
        *args: typing.Any,
        **kwargs: typing.Any,
    ) -> None:
        super().__init__(*args, **kwargs)
        self.serialize_extended_info = False

    def __str__(self) -> str:
        if self.ErrorDetails and self.ErrorDetails.Message:
            return self.ErrorDetails.Message

        else:
            return self.Exception.Message

    @classmethod
    def FromPSObjectForRemoting(
        cls,
        obj: PSObject,
        **kwargs: typing.Any,
    ) -> "ErrorRecord":
        category_info = ErrorCategoryInfo(
            Category=ErrorCategory(obj.ErrorCategory_Category),
            Activity=obj.ErrorCategory_Activity,
            Reason=obj.ErrorCategory_Reason,
            TargetName=obj.ErrorCategory_TargetName,
            TargetType=obj.ErrorCategory_TargetType,
        )
        category_info.PSObject.to_string = obj.ErrorCategory_Message

        error_details = None
        error_details_message = getattr(obj, "ErrorDetails_Message", None)
        error_details_action = getattr(obj, "ErrorDetails_RecommendedAction", None)
        if error_details_message or error_details_action:
            error_details = ErrorDetails(
                Message=error_details_message,
                RecommendedAction=error_details_action,
            )

        # Technically PowerShell wraps the exception in a RemoteException class which contains
        # 'SerializedRemoteException' and 'SerializedRemoteInvocationInfo'. To make things simple we just use the
        # serialized exception as the actual Exception value and add the invocation info to that.
        add_note_property(obj.Exception, "SerializedRemoteInvocationInfo", obj.InvocationInfo, force=True)

        invocation_info = None
        pipeline_iteration_info = None
        if obj.SerializeExtendedInfo:
            pipeline_iteration_info = obj.PipelineIterationInfo
            invocation_info = _invocation_info_from_ps_object(obj)

        record = cls(
            Exception=obj.Exception,
            TargetObject=obj.TargetObject,
            FullyQualifiedErrorId=obj.FullyQualifiedErrorId,
            InvocationInfo=invocation_info,
            CategoryInfo=category_info,
            ErrorDetails=error_details,
            PipelineIterationInfo=pipeline_iteration_info,
            ScriptStackTrace=getattr(obj, "ErrorDetails_ScriptStackTrace", None),
        )
        record.serialize_extended_info = obj.SerializeExtendedInfo
        return record

    @classmethod
    def ToPSObjectForRemoting(
        cls,
        instance: "ErrorRecord",
        **kwargs: typing.Any,
    ) -> PSObject:
        obj = PSObject()
        add_note_property(obj, "Exception", instance.Exception)
        add_note_property(obj, "TargetObject", instance.TargetObject)
        add_note_property(obj, "FullyQualifiedErrorId", instance.FullyQualifiedErrorId)
        add_note_property(obj, "InvocationInfo", instance.InvocationInfo)
        add_note_property(obj, "ErrorCategory_Category", instance.CategoryInfo.Category.value)
        add_note_property(obj, "ErrorCategory_Activity", instance.CategoryInfo.Activity)
        add_note_property(obj, "ErrorCategory_Reason", instance.CategoryInfo.Reason)
        add_note_property(obj, "ErrorCategory_TargetName", instance.CategoryInfo.TargetName)
        add_note_property(obj, "ErrorCategory_TargetType", instance.CategoryInfo.TargetType)
        add_note_property(obj, "ErrorCategory_Message", str(instance.CategoryInfo))

        if instance.ErrorDetails:
            add_note_property(obj, "ErrorDetails_Message", instance.ErrorDetails.Message)
            add_note_property(obj, "ErrorDetails_RecommendedAction", instance.ErrorDetails.RecommendedAction)

        if instance.ScriptStackTrace:
            add_note_property(obj, "ErrorDetails_ScriptStackTrace", instance.ScriptStackTrace)

        if instance.serialize_extended_info and instance.InvocationInfo:
            add_note_property(obj, "SerializeExtendedInfo", True)
            _invocation_info_to_ps_object(instance.InvocationInfo, obj)
            add_note_property(obj, "PipelineIterationInfo", instance.PipelineIterationInfo)

        else:
            add_note_property(obj, "SerializeExtendedInfo", False)

        return obj


@PSType(
    type_names=[
        "System.Management.Automation.InformationalRecord",
    ],
    adapted_properties=[
        PSNoteProperty("Message", ps_type=PSString),
        PSNoteProperty("InvocationInfo", ps_type=InvocationInfo),
        PSNoteProperty("PipelineIterationInfo", ps_type=PSList),
    ],
)
class InformationalRecord(PSObject):
    """PowerShell InformationalRecord.

    InformationalRecord (that is Debug, Warning, or Verbose) is a structure
    that contains additional information that a pipeline can output in addition
    to the regular data output. It is documented in PSRP under
    `[MS-PSRP] 2.2.3.16 InformationalRecord`_. The invocation specific
    properties are documented under `[MS-PSRP] 2.2.3.15.1 InvocationInfo`_.
    This also represents the
    `System.Management.Automation.InformationalRecord`_ .NET type.

    Args:
        Message: The message writen by the command that created this record.
        InvocationInfo: The invocation info of the command that created this
            record.
        PipelineIterationInfo: The status of the pipeline when this record was
            created.

    .. _[MS-PSRP] 2.2.3.16 InformationalRecord:
        https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-psrp/97cad2dc-c34a-4db6-bfa1-cbf196853937

    .. _[MS-PSRP] 2.2.3.15.1 InvocationInfo:
        https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-psrp/000363b7-e2f9-4a34-94f5-d540a15aee7b

    .. _System.Management.Automation.InformationalRecord:
        https://docs.microsoft.com/en-us/dotnet/api/system.management.automation.informationalrecord
    """

    def __init__(
        self,
        *args: typing.Any,
        **kwargs: typing.Any,
    ) -> None:
        super().__init__(*args, **kwargs)
        self.serialize_extended_info = False

    @classmethod
    def FromPSObjectForRemoting(
        cls,
        obj: PSObject,
        **kwargs: typing.Any,
    ) -> "InformationalRecord":
        invocation_info = None
        pipeline_iteration_info = None
        if obj.InformationalRecord_SerializeInvocationInfo:
            pipeline_iteration_info = obj.InformationalRecord_PipelineIterationInfo
            invocation_info = _invocation_info_from_ps_object(obj)

        record = cls(
            Message=obj.InformationalRecord_Message,
            InvocationInfo=invocation_info,
            PipelineIterationInfo=pipeline_iteration_info,
        )
        record.serialize_extended_info = obj.InformationalRecord_SerializeInvocationInfo

        return record

    @classmethod
    def ToPSObjectForRemoting(
        cls,
        instance: "ErrorRecord",
        **kwargs: typing.Any,
    ) -> PSObject:
        obj = PSObject()
        add_note_property(obj, "InformationalRecord_Message", instance.Message)

        if instance.serialize_extended_info and instance.InvocationInfo:
            add_note_property(obj, "InformationalRecord_SerializeInvocationInfo", True)
            _invocation_info_to_ps_object(instance.InvocationInfo, obj)
            add_note_property(obj, "InformationalRecord_PipelineIterationInfo", instance.PipelineIterationInfo)

        else:
            add_note_property(obj, "InformationalRecord_SerializeInvocationInfo", False)

        return obj


@PSType(["System.Management.Automation.DebugRecord"])
class DebugRecord(InformationalRecord):
    """DebugRecord.

    A debug record in the PSInformationalBuffers. This represents the
    `System.Management.Automation.DebugRecord`_ .NET type.

    .. _System.Management.Automation.DebugRecord:
        https://docs.microsoft.com/en-us/dotnet/api/system.management.automation.debugrecord?view=powershellsdk-7.0.0
    """


@PSType(["System.Management.Automation.VerboseRecord"])
class VerboseRecord(InformationalRecord):
    """VerboseRecord.

    A verbose record in the PSInformationalBuffers. This represents the
    `System.Management.Automation.VerboseRecord`_ .NET type.

    .. _System.Management.Automation.VerboseRecord:
        https://docs.microsoft.com/en-us/dotnet/api/system.management.automation.verboserecord?view=powershellsdk-7.0.0
    """


@PSType(["System.Management.Automation.WarningRecord"])
class WarningRecord(InformationalRecord):
    """WarningRecord.

    A warning record in the PSInformationalBuffers. This represents the
    `System.Management.Automation.WarningRecord`_ .NET type.

    .. _System.Management.Automation.WarningRecord:
        https://docs.microsoft.com/en-us/dotnet/api/system.management.automation.warningrecord?view=powershellsdk-7.0.0
    """


@PSType(
    type_names=["System.Management.Automation.ProgressRecord"],
    adapted_properties=[
        PSNoteProperty("ActivityId", mandatory=True, ps_type=PSInt),
        PSNoteProperty("Activity", mandatory=True, ps_type=PSString),
        PSNoteProperty("StatusDescription", mandatory=True, ps_type=PSString),
        PSNoteProperty("CurrentOperation", ps_type=PSString),
        PSNoteProperty("ParentActivityId", value=-1, ps_type=PSInt),
        PSNoteProperty("PercentComplete", value=-1, ps_type=PSInt),
        PSNoteProperty("RecordType", value=ProgressRecordType.Processing, ps_type=ProgressRecordType),
        PSNoteProperty("SecondsRemaining", value=-1, ps_type=PSInt),
    ],
    tag="PR",
)
class ProgressRecord(PSObject):
    """ProgressRecord.

    Defines a data structure used to represent the status of an ongoing
    operation at a point in time. This represents the
    `System.Management.Automation.ProgressRecord`_ .NET type.

    Args:
        Activity: Description of the activity.
        ActivityId: Id of the activity, used as a key for the linking of
            subordinate activities.

        CurrentOperation: Current operation of the many required to accomplish
            the activity, e.g. "copying foo.txt".
        ParentActivityId: Id of the activity for which this is a subordinate.
        PercentComplete: Percentage of total work for the activity that is
            completed.
        RecordType: Type of record represented by this instance.
        SecondsRemaining: Estimated time remaining until the activity is
            complete.
        StatusDescription: Current status of the operation, e.g. "35 of 50
            items copied".

    .. _System.Management.Automation.ProgressRecord:
        https://docs.microsoft.com/en-us/dotnet/api/system.management.automation.progressrecord?view=powershellsdk-7.0.0
    """


@PSType(
    ["System.Management.Automation.InformationRecord"],
    adapted_properties=[
        PSNoteProperty("MessageData"),
        PSNoteProperty("Source", ps_type=PSString),
        PSNoteProperty("TimeGenerated", ps_type=PSDateTime),
        PSNoteProperty("Tags", ps_type=PSList),
        PSNoteProperty("User", ps_type=PSString),
        PSNoteProperty("Computer", ps_type=PSString),
        PSNoteProperty("ProcessId", ps_type=PSUInt),
        PSNoteProperty("NativeThreadId", ps_type=PSUInt),
        PSNoteProperty("ManagedThreadId", ps_type=PSUInt),
    ],
)
class InformationRecord(PSObject):
    """InformationRecord.

    Defines a data structure used to represent informational context destined
    for the host or user. This represents the
    `System.Management.Automation.InformationRecord`_ .NET type.

    Args:
        MessageData: The message data for the informational record.
        Source: The source of the information record (script path, function
            name, etc.).
        TimeGenerated: The time the informational record was generated.
        Tags: The tags associated with this informational record.
        User: THe user that generated the informational record.
        Computer: THe computer that generated the informational record.
        ProcessId: The process that generated the informational record.
        NativeThreadId: The native thread that generated the informational
            record.
        ManagedThreadId: The managed thread that generated the informational
            record.

    .. _System.Management.Automation.InformationRecord:
        https://docs.microsoft.com/en-us/dotnet/api/system.management.automation.informationrecord?view=powershellsdk-7.0.0
    """


@PSType(type_names=["System.Management.Automation.PSPrimitiveDictionary"])
class PSPrimitiveDictionary(PSDict):
    """Primitive Dictionary.

    A primitive dictionary represents a dictionary which contains only objects
    that are primitive types. While Python does not place any limitations on
    the types this object can contain, trying to serialize a
    PSPrimitiveDictionary with complex types to PowerShell will fail. The types
    that are allowed can be found at `[MS-PSRP] 2.2.3.18 Primitive Dictionary`_.

    .. _[MS-PSRP] 2.2.3.18 Primitive Dictionary:
        https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-psrp/7779aa42-6927-4225-b31c-2771fd869546
    """


@PSType(
    type_names=["Selected.Microsoft.PowerShell.Commands.GenericMeasureInfo"],
    extended_properties=[
        PSNoteProperty("Count", mandatory=True, ps_type=PSInt),
    ],
)
class CommandMetadataCount(PSCustomObject):
    """CommandMetadataCount.

    Special data type used by the command metadata messages. It is documented
    in PSRP under `[MS-PSRP] 2.2.3.21 CommandMetadataCount`_.

    Args:
        Count: The count.

    .. _[MS-PSRP] 2.2.3.21 CommandMetadataCount:
        https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-psrp/4647da0c-18e6-496c-9d9e-c669d40dc1db
    """


@PSType(
    type_names=["System.Management.Automation.PSCredential"],
    adapted_properties=[
        PSNoteProperty("UserName", mandatory=True, ps_type=PSString),
        PSNoteProperty("Password", mandatory=True, ps_type=PSSecureString),
    ],
)
class PSCredential(PSObject):
    """PSCredential.

    Represents a username and a password. It is documented in PSRP under
    `[MS-PSRP] 2.2.3.25 PSCredential`_. It also represents the
    `System.Management.Automation.PSCredential`_ .NET type.

    Note:
        To be able to serialize this object, a session key is exchanged between
        the host and the peer. If the peer does not support the session key
        exchange then this cannot be serialized.

    Args:
        UserName: The username for the credential.
        Password: The password for the credential.

    .. _[MS-PSRP] 2.2.3.25 PSCredential:
        https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-psrp/a7c91a93-ee59-4af0-8a67-a9361af9870e

    .. _System.Management.Automation.PSCredential:
        https://docs.microsoft.com/en-us/dotnet/api/system.management.automation.pscredential
    """


@PSType(
    type_names=[
        "System.Exception",
    ],
    adapted_properties=[
        PSNoteProperty("Message", mandatory=True, ps_type=PSString),
        PSNoteProperty("Data", ps_type=PSDict),
        PSNoteProperty("HelpLink", ps_type=PSString),
        PSNoteProperty("HResult", ps_type=PSInt),
        PSNoteProperty("InnerException"),
        PSNoteProperty("Source", ps_type=PSString),
        PSNoteProperty("StackTrace", ps_type=PSString),
        PSNoteProperty("TargetSite", ps_type=PSString),
    ],
)
class NETException(PSObject):
    """.NET Exception.

    Represents a .NET `System.Exception`_ type. It isn't documented in
    MS-PSRP but is used when creating an ErrorRecord or just as a base of
    another exception type.

    Args:
        Message: Message that describes the current exception.
        Data: User defined information about the exception.
        HelpLink: A link to the help file associated with this exception.
        HResult: A coded numerical value that is assigned to a specific exception.
        InnerException: Exception instance that caused the current exception.
        Source: Name of the application or the object that causes the error.
        StackTrace: String representation of the immediate frames on the call stack.
        TargetSite: Method that throws the current exception.

    .. _System.Exception:
        https://docs.microsoft.com/en-us/dotnet/api/system.exception?view=net-5.0
    """



# -*- coding: utf-8 -*-
# Copyright: (c) 2021, Jordan Borean (@jborean93) <jborean93@gmail.com>
# MIT License (see LICENSE or https://opensource.org/licenses/MIT)

import typing

from psrpcore.types import PSInvocationState, PSVersion, RunspacePoolState

T = typing.TypeVar("T")


class PSRPCoreError(Exception):
    """Base error for any PSRP errors."""


class MissingCipherError(PSRPCoreError):
    """Trying to (de)serialize a Secure String but no cipher was provided."""

    @property
    def message(self) -> str:
        return "Cannot (de)serialize a secure string without an exchanged session key"

    def __str__(self) -> str:
        return self.message


class _InvalidState(PSRPCoreError, typing.Generic[T]):
    _STATE_OBJ: typing.Optional[str] = None

    def __init__(
        self,
        action: str,
        current_state: T,
        expected_states: typing.List[T],
    ) -> None:
        self.action = action
        self.current_state = current_state
        self.expected_states = expected_states

    @property
    def message(self) -> str:
        expected_states = ", ".join(str(s) for s in self.expected_states)
        return (
            f"{self._STATE_OBJ} state must be one of '{expected_states}' to {self.action}, current state is "
            f"{self.current_state!s}"
        )

    def __str__(self) -> str:
        return self.message


class InvalidRunspacePoolState(_InvalidState[RunspacePoolState]):
    """The Runspace Pool is not in the required state."""

    _STATE_OBJ = "Runspace Pool"


class InvalidPipelineState(_InvalidState[PSInvocationState]):
    """The Pipeline is not in the required state."""

    _STATE_OBJ = "Pipeline"


class InvalidProtocolVersion(PSRPCoreError):
    """The protocolversion of the peer does not meet the required version."""

    def __init__(
        self,
        action: str,
        current_version: PSVersion,
        required_version: PSVersion,
    ):
        self.action = action
        self.current_version = current_version
        self.required_version = required_version

    @property
    def message(self) -> str:
        return (
            f"{self.action} requires a protocol version of {self.required_version}, current version is "
            f"{self.current_version}"
        )

    def __str__(self) -> str:
        return self.message



# -*- coding: utf-8 -*-
# Copyright: (c) 2021, Jordan Borean (@jborean93) <jborean93@gmail.com>
# MIT License (see LICENSE or https://opensource.org/licenses/MIT)

import xml.etree.ElementTree as ElementTree

import psrpcore.types._host as host
from psrpcore.types import ConsoleColor, PSChar, PSObject

from ..conftest import assert_xml_diff, deserialize, serialize


def test_buffer_cell():
    value = host.BufferCell(97, ConsoleColor.White, ConsoleColor.Black, host.BufferCellType.Leading)
    assert value.Character == PSChar("a")
    assert value.ForegroundColor == ConsoleColor.White
    assert value.BackgroundColor == ConsoleColor.Black
    assert value.BufferCellType == host.BufferCellType.Leading

    element = serialize(value)
    actual = ElementTree.tostring(element, encoding="utf-8").decode()
    expected = (
        '<Obj RefId="0">'
        '<TN RefId="0">'
        "<T>System.Management.Automation.Host.BufferCell</T>"
        "<T>System.ValueType</T>"
        "<T>System.Object</T>"
        "</TN>"
        "<Props>"
        '<C N="Character">97</C>'
        '<Obj RefId="1" N="ForegroundColor">'
        "<I32>15</I32>"
        '<TN RefId="1">'
        "<T>System.ConsoleColor</T>"
        "<T>System.Enum</T>"
        "<T>System.ValueType</T>"
        "<T>System.Object</T>"
        "</TN>"
        "<ToString>White</ToString>"
        "</Obj>"
        '<Obj RefId="2" N="BackgroundColor"><I32>0</I32><TNRef RefId="1" /><ToString>Black</ToString></Obj>'
        '<Obj RefId="3" N="BufferCellType">'
        "<I32>1</I32>"
        '<TN RefId="2">'
        "<T>System.Management.Automation.Host.ControlKeyStates</T>"
        "<T>System.Enum</T>"
        "<T>System.ValueType</T>"
        "<T>System.Object</T>"
        "</TN>"
        "<ToString>Leading</ToString>"
        "</Obj>"
        "</Props>"
        "</Obj>"
    )
    assert_xml_diff(actual, expected)

    value = deserialize(element)
    assert isinstance(value, PSObject)
    assert isinstance(value, host.BufferCell)
    assert value.Character == PSChar("a")
    assert value.ForegroundColor == ConsoleColor.White
    assert value.BackgroundColor == ConsoleColor.Black
    assert value.BufferCellType == host.BufferCellType.Leading

    assert host.BufferCell.FromPSObjectForRemoting(value) is value


def test_choice_description():
    value = host.ChoiceDescription("label", "help")
    assert value.Label == "label"
    assert value.HelpMessage == "help"

    element = serialize(value)
    actual = ElementTree.tostring(element, encoding="utf-8").decode()
    assert_xml_diff(
        actual,
        '<Obj RefId="0">'
        '<TN RefId="0">'
        "<T>System.Management.Automation.Host.ChoiceDescription</T>"
        "<T>System.Object</T>"
        "</TN>"
        "<Props>"
        '<S N="Label">label</S>'
        '<S N="HelpMessage">help</S>'
        "</Props>"
        "</Obj>",
    )

    value = deserialize(element)
    assert isinstance(value, PSObject)
    assert isinstance(value, host.ChoiceDescription)
    assert value.Label == "label"
    assert value.HelpMessage == "help"

    assert host.ChoiceDescription.FromPSObjectForRemoting(value) is value


def test_coordinates():
    value = host.Coordinates(10, 412)
    assert value.X == 10
    assert value.Y == 412

    element = serialize(value)
    actual = ElementTree.tostring(element, encoding="utf-8").decode()
    assert_xml_diff(
        actual,
        '<Obj RefId="0">'
        '<TN RefId="0">'
        "<T>System.Management.Automation.Host.Coordinates</T>"
        "<T>System.ValueType</T>"
        "<T>System.Object</T>"
        "</TN>"
        "<Props>"
        '<I32 N="X">10</I32>'
        '<I32 N="Y">412</I32>'
        "</Props>"
        "</Obj>",
    )

    value = deserialize(element)
    assert isinstance(value, PSObject)
    assert isinstance(value, host.Coordinates)
    assert value.X == 10
    assert value.Y == 412

    assert host.Coordinates.FromPSObjectForRemoting(value) is value


def test_field_description():
    value = host.FieldDescription("name")
    assert value.Name == "name"
    assert value.ParameterTypeName is None
    assert value.ParameterTypeFullName is None
    assert value.ParameterAssemblyFullName is None
    assert value.Label == ""
    assert value.HelpMessage == ""
    assert value.IsMandatory is True
    assert value.DefaultValue is None
    assert value.Attributes is None

    element = serialize(value)
    actual = ElementTree.tostring(element, encoding="utf-8").decode()
    assert_xml_diff(
        actual,
        '<Obj RefId="0">'
        '<TN RefId="0">'
        "<T>System.Management.Automation.Host.FieldDescription</T>"
        "<T>System.Object</T>"
        "</TN>"
        "<Props>"
        '<S N="Name">name</S>'
        '<Nil N="ParameterTypeName" />'
        '<Nil N="ParameterTypeFullName" />'
        '<Nil N="ParameterAssemblyFullName" />'
        '<S N="Label" />'
        '<S N="HelpMessage" />'
        '<B N="IsMandatory">true</B>'
        '<Nil N="DefaultValue" />'
        '<Nil N="Attributes" />'
        "</Props>"
        "</Obj>",
    )

    value = deserialize(element)
    assert isinstance(value, PSObject)
    assert isinstance(value, host.FieldDescription)
    assert value.ParameterTypeName is None
    assert value.ParameterTypeFullName is None
    assert value.ParameterAssemblyFullName is None
    assert value.Label == ""
    assert value.HelpMessage == ""
    assert value.IsMandatory is True
    assert value.DefaultValue is None
    assert value.Attributes is None

    assert host.FieldDescription.FromPSObjectForRemoting(value) is value


def test_key_info():
    value = host.KeyInfo(10, 97, host.ControlKeyStates.EnhancedKey, False)
    assert value.VirtualKeyCode == 10
    assert value.Character == PSChar("a")
    assert value.ControlKeyState == host.ControlKeyStates.EnhancedKey
    assert value.KeyDown is False

    element = serialize(value)
    actual = ElementTree.tostring(element, encoding="utf-8").decode()
    assert_xml_diff(
        actual,
        (
            '<Obj RefId="0">'
            '<TN RefId="0">'
            "<T>System.Management.Automation.Host.KeyInfo</T>"
            "<T>System.Object</T>"
            "</TN>"
            "<Props>"
            '<I32 N="VirtualKeyCode">10</I32>'
            '<C N="Character">97</C>'
            '<Obj RefId="1" N="ControlKeyState">'
            "<I32>256</I32>"
            '<TN RefId="1">'
            "<T>System.Management.Automation.Host.ControlKeyStates</T>"
            "<T>System.Enum</T>"
            "<T>System.ValueType</T>"
            "<T>System.Object</T>"
            "</TN>"
            "<ToString>EnhancedKey</ToString>"
            "</Obj>"
            '<B N="KeyDown">false</B>'
            "</Props>"
            "</Obj>"
        ),
    )

    value = deserialize(element)
    assert isinstance(value, PSObject)
    assert isinstance(value, host.KeyInfo)
    assert value.VirtualKeyCode == 10
    assert value.Character == PSChar("a")
    assert value.ControlKeyState == host.ControlKeyStates.EnhancedKey
    assert value.KeyDown is False

    assert host.KeyInfo.FromPSObjectForRemoting(value) is value


def test_rectangle():
    value = host.Rectangle(1, 2, 3, 4)
    assert value.Left == 1
    assert value.Top == 2
    assert value.Right == 3
    assert value.Bottom == 4

    element = serialize(value)
    actual = ElementTree.tostring(element, encoding="utf-8").decode()
    assert_xml_diff(
        actual,
        '<Obj RefId="0">'
        '<TN RefId="0">'
        "<T>System.Management.Automation.Host.Rectangle</T>"
        "<T>System.ValueType</T>"
        "<T>System.Object</T>"
        "</TN>"
        "<Props>"
        '<I32 N="Left">1</I32>'
        '<I32 N="Top">2</I32>'
        '<I32 N="Right">3</I32>'
        '<I32 N="Bottom">4</I32>'
        "</Props>"
        "</Obj>",
    )

    value = deserialize(element)
    assert isinstance(value, PSObject)
    assert isinstance(value, host.Rectangle)
    assert value.Left == 1
    assert value.Top == 2
    assert value.Right == 3
    assert value.Bottom == 4

    assert host.Rectangle.FromPSObjectForRemoting(value) is value


def test_size():
    value = host.Size(10, 412)
    assert value.Width == 10
    assert value.Height == 412

    element = serialize(value)
    actual = ElementTree.tostring(element, encoding="utf-8").decode()
    assert_xml_diff(
        actual,
        '<Obj RefId="0">'
        '<TN RefId="0">'
        "<T>System.Management.Automation.Host.Size</T>"
        "<T>System.ValueType</T>"
        "<T>System.Object</T>"
        "</TN>"
        "<Props>"
        '<I32 N="Width">10</I32>'
        '<I32 N="Height">412</I32>'
        "</Props>"
        "</Obj>",
    )

    value = deserialize(element)
    assert isinstance(value, PSObject)
    assert isinstance(value, host.Size)
    assert value.Width == 10
    assert value.Height == 412

    assert host.Size.FromPSObjectForRemoting(value) is value



# -*- coding: utf-8 -*-
# Copyright: (c) 2021, Jordan Borean (@jborean93) <jborean93@gmail.com>
# MIT License (see LICENSE or https://opensource.org/licenses/MIT)

import typing

from psrpcore._command import Command
from psrpcore._payload import dict_to_psobject
from psrpcore.types import (
    ApartmentState,
    CommandTypes,
    CreatePipeline,
    GetCommandMetadata,
    HostInfo,
    PSObject,
    RemoteStreamOptions,
)


class PowerShell(PSObject):
    """PowerShell Pipeline.

    This implements the PowerShell pipeline specific methods used to invoke a
    PowerShell pipeline.

    Args:
        add_to_history: Whether to add the pipeline to the history field of the
            runspace.
        apartment_state: The apartment state of the thread that executes the
            pipeline.
        history: The value to use as a historial reference of the pipeline.
        host: The host information to use when executing the pipeline.
        is_nested: Whether the pipeline is nested in another pipeline or not.
        no_input: Whether there is any data to be input into the pipeline.
        remote_stream_options: Whether to add invocation info the the PowerShell
            streams or not.
        redirect_shell_error_to_out: Redirects the global error output pipe to
            the commands error output pipe.
    """

    def __init__(
        self,
        add_to_history: bool = False,
        apartment_state: ApartmentState = None,
        history: typing.Optional[str] = None,
        host: typing.Optional[HostInfo] = None,
        is_nested: bool = False,
        no_input: bool = True,
        remote_stream_options: RemoteStreamOptions = RemoteStreamOptions.none,
        redirect_shell_error_to_out: bool = True,
    ) -> None:
        self.add_to_history = add_to_history
        self.apartment_state = apartment_state
        self.commands: typing.List[Command] = []
        self.history = history
        self.host = host or HostInfo()
        self.is_nested = is_nested
        self.no_input = no_input
        self.remote_stream_options = remote_stream_options
        self.redirect_shell_error_to_out = redirect_shell_error_to_out

    def __repr__(self) -> str:
        return (
            f"<{type(self).__name__} add_to_history={self.add_to_history} apartment_state={self.apartment_state!r} "
            f"commands={self.commands!r} history={self.history!r} host={self.host!r} is_nested={self.is_nested} "
            f"no_input={self.no_input} remote_stream_options={self.remote_stream_options!r} "
            f"redirect_shell_error_to_out={self.redirect_shell_error_to_out!r}>"
        )

    @classmethod
    def FromPSObjectForRemoting(
        cls,
        value: "PSObject",
        **kwargs: typing.Any,
    ) -> "PowerShell":
        powershell = value.PowerShell

        pipeline = PowerShell(
            add_to_history=value.AddToHistory,
            apartment_state=value.ApartmentState,
            history=powershell.History,
            host=HostInfo.FromPSObjectForRemoting(value.HostInfo),
            is_nested=value.IsNested,
            no_input=value.NoInput,
            remote_stream_options=value.RemoteStreamOptions,
            redirect_shell_error_to_out=powershell.RedirectShellErrorOutputPipe,
        )

        extra_cmds = getattr(powershell, "ExtraCmds", None)
        if extra_cmds is not None:
            commands = [c.Cmds for c in extra_cmds]
        else:
            commands = [powershell.Cmds]

        for statements in commands:
            for raw_cmd in statements:
                cmd = Command.FromPSObjectForRemoting(raw_cmd)
                pipeline.commands.append(cmd)

            pipeline.commands[-1].end_of_statement = True

        return pipeline

    @classmethod
    def ToPSObjectForRemoting(
        cls,
        instance: "PowerShell",
        **kwargs: typing.Any,
    ) -> CreatePipeline:
        if not instance.commands:
            raise ValueError("A command is required to invoke a PowerShell pipeline.")

        extra_cmds: typing.List[typing.List[PSObject]] = [[]]
        for cmd in instance.commands:
            extra_cmds[-1].append(cmd)
            if cmd.end_of_statement:
                extra_cmds.append([])
        cmds = extra_cmds[0]

        # MS-PSRP 2.2.3.11 Pipeline
        # https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-psrp/82a8d1c6-4560-4e68-bfd0-a63c36d6a199
        pipeline_kwargs = {
            "Cmds": cmds,
            "IsNested": instance.is_nested,
            "History": instance.history,
            "RedirectShellErrorOutputPipe": instance.redirect_shell_error_to_out,
        }

        if len(extra_cmds) > 1:
            # This isn't documented in MS-PSRP but this is how PowerShell batches multiple statements in 1 pipeline.
            # TODO: ExtraCmds may not work with protocol <=2.1.
            pipeline_kwargs["ExtraCmds"] = [dict_to_psobject(Cmds=s) for s in extra_cmds]

        return CreatePipeline(
            NoInput=instance.no_input,
            ApartmentState=instance.apartment_state,
            RemoteStreamOptions=instance.remote_stream_options,
            AddToHistory=instance.add_to_history,
            HostInfo=instance.host,
            PowerShell=dict_to_psobject(**pipeline_kwargs),
            IsNested=instance.is_nested,
        )


class GetMetadata(PSObject):
    """Get Command Metadata Pipeline.

    This implements the GetCommandMetadata pipeline specific methods used to get
    command metadata information.

    Args:
        name: List of command names to get the metadata for. Uses ``*`` as a
            wildcard.
        command_type: The type of commands to filter by.
        namespace: Wildcard patterns describbing the command namespace to filter
            by.
        arguments: Extra arguments passed to the higher-layer above PSRP.
    """

    def __init__(
        self,
        name: typing.Union[str, typing.List[str]],
        command_type: CommandTypes = CommandTypes.All,
        namespace: typing.Optional[typing.List[str]] = None,
        arguments: typing.Optional[typing.List[typing.Any]] = None,
    ) -> None:
        if not isinstance(name, list):
            name = [name]
        self.name = name
        self.command_type = command_type
        self.namespace = namespace
        self.arguments = arguments

    def __repr__(self) -> str:
        return (
            f"<{type(self).__name__} name={self.name!r} command_type={self.command_type!r} "
            f"namespace={self.namespace!r} arguments={self.arguments!r}>"
        )

    @classmethod
    def FromPSObjectForRemoting(
        cls,
        value: PSObject,
        **kwargs: typing.Any,
    ) -> "GetMetadata":
        return GetMetadata(
            name=value.Name,
            command_type=value.CommandType,
            namespace=value.Namespace,
            arguments=value.ArgumentList,
        )

    @classmethod
    def ToPSObjectForRemoting(
        cls,
        instance: "GetMetadata",
        **kwargs: typing.Any,
    ) -> GetCommandMetadata:
        return GetCommandMetadata(
            Name=instance.name,
            CommandType=instance.command_type,
            Namespace=instance.namespace,
            ArgumentList=instance.arguments,
        )



# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
import os
import sys
import typing

from sphinx.application import Sphinx

sys.path.insert(0, os.path.abspath(".."))

import re

from psrpcore.types import PSEnumBase, PSFlagBase, PSObject
from psrpcore.types._base import _UnsetValue

_PARAM_PATTERN = re.compile(r":param (\w*):")

# -- Project information -----------------------------------------------------

project = "psrpcore"
copyright = "2021, Jordan Borean"
author = "Jordan Borean"


# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    "myst_parser",
    "sphinx_rtd_theme",
    "sphinx.ext.autodoc",
    "sphinx.ext.coverage",
    "sphinx.ext.napoleon",
    "sphinxcontrib.apidoc",
]

apidoc_module_dir = "../src/psrpcore"
apidoc_output_dir = "source"

# Add any paths that contain templates here, relative to this directory.
templates_path = [
    "_templates",
]

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = [
    "_build",
    "Thumbs.db",
    ".DS_Store",
]

source_suffix = {
    ".rst": "restructuredtext",
    ".md": "markdown",
}


# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = "sphinx_rtd_theme"

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ["_static"]


def autodoc_process_docstring(
    app: Sphinx,
    what: str,
    name: str,
    obj: typing.Any,
    options: typing.Dict[str, typing.Any],
    lines: typing.List[str],
) -> None:
    """Auto inserts param type details for PSObjects."""
    if what != "class" or not issubclass(obj, PSObject):
        return

    prop_entries = {p.name: p for p in obj.PSObject.adapted_properties}
    prop_entries.update({p.name: p for p in obj.PSObject.extended_properties})

    total_lines = len(lines)
    insertions = []
    for idx in range(total_lines):
        line = lines[idx]
        param_match = _PARAM_PATTERN.match(line)
        if not param_match:
            continue

        param_name = param_match.group(1)

        # Need to get the next line that isn't a continuation of the param docs
        # e.g. has no leading spaces.
        next_line = ""
        while idx + 1 != total_lines:
            idx += 1
            next_line = lines[idx]
            if not next_line.startswith(" "):
                break

        if next_line.startswith(f":type {param_name}:"):
            continue

        insertions.append((idx, param_name))

    insertion_offset = 0
    for idx, param_name in insertions:
        prop = prop_entries.get(param_name)
        if not prop:
            continue

        param_type = prop.ps_type if prop.ps_type is not None else PSObject
        if param_type.__module__ == "builtins":
            param_type_str = param_type.__name__

        else:
            param_type_str = f"{param_type.__module__}.{param_type.__name__}"

        type_line = f":type {param_name}: :obj:`{param_type.__name__} " f"<{param_type_str}>`"
        lines.insert(idx + insertion_offset, type_line)
        insertion_offset += 1


def autodoc_process_signature(
    app: Sphinx,
    what: str,
    name: str,
    obj: typing.Any,
    options: typing.Dict[str, typing.Any],
    signature: typing.Optional[str],
    return_annotations: typing.Optional[str],
) -> typing.Optional[typing.Tuple[typing.Optional[str], typing.Optional[str]]]:
    """Auto format PSObject signatures to document proper args/kwargs."""
    if what != "class":
        return None

    # Make sure PS enum types have the proper enum signature.
    if issubclass(obj, (PSEnumBase, PSFlagBase)):
        return "(value)", return_annotations

    # Make sure PSObject classes that use __init__ from PSObject define the
    # proper signature.
    if not isinstance(obj, PSObject) and obj.__init__ != PSObject.__init__:
        return None

    prop_entries = {p.name: p for p in obj.PSObject.adapted_properties}
    prop_entries.update({p.name: p for p in obj.PSObject.extended_properties})
    kwargs = []
    for prop in prop_entries.values():
        ps_type = prop.ps_type if prop.ps_type is not None else PSObject

        if ps_type.__module__ == "builtins":
            ps_type_str = ps_type.__name__

        else:
            ps_type_str = ps_type.__name__

        if prop.mandatory:
            entry = f"{prop.name}: {ps_type_str}"

        else:
            default_value = prop._value
            if prop._value == _UnsetValue:
                default_value = None

            entry = f"{prop.name}: Optional[{ps_type_str}] = {default_value!r}"

        kwargs.append(entry)

    signature = f'({", ".join(kwargs)})' if kwargs else None

    return signature, return_annotations


def autodoc_skip_member_handler(
    app: Sphinx,
    what: str,
    name: str,
    obj: typing.Any,
    skip: bool,
    options: typing.Dict[str, typing.Any],
) -> bool:
    a = ""
    return (
        skip
        # This is a metadata class for all PSObject types, end users don't need
        # to know about this for every single class in the codebase.
        or (name == "PSObject" and (obj != PSObject and obj != PSObject.PSObject))
        or name
        in [
            # These 2 methods are used to customize the (de)serialization process
            # for a specific class. They are not for public use and should not be
            # documented.
            "FromPSObjectForRemoting",
            "ToPSObjectForRemoting",
        ]
    )


def setup(
    app: Sphinx,
) -> None:

    app.connect("autodoc-process-docstring", autodoc_process_docstring)
    app.connect("autodoc-process-signature", autodoc_process_signature)
    app.connect("autodoc-skip-member", autodoc_skip_member_handler)



# -*- coding: utf-8 -*-
# Copyright: (c) 2021, Jordan Borean (@jborean93) <jborean93@gmail.com>
# MIT License (see LICENSE or https://opensource.org/licenses/MIT)

from psrpcore import _exceptions as exp
from psrpcore.types import PSInvocationState, PSVersion, RunspacePoolState


def test_missing_cipher():
    missing_cipher = exp.MissingCipherError()
    str(missing_cipher) == "Cannot (de)serialize a secure string without an exchanged session key"


def test_invalid_runspace_pool_state():
    invalid = exp.InvalidRunspacePoolState("open pool", RunspacePoolState.BeforeOpen, [RunspacePoolState.Opened])
    assert str(invalid) == (
        "Runspace Pool state must be one of 'RunspacePoolState.Opened' to open pool, "
        "current state is RunspacePoolState.BeforeOpen"
    )


def test_invalid_pipeline_state():
    invalid = exp.InvalidPipelineState(
        "open pipe", PSInvocationState.Completed, [PSInvocationState.Running, PSInvocationState.Stopped]
    )
    assert str(invalid) == (
        "Pipeline state must be one of 'PSInvocationState.Running, PSInvocationState.Stopped' to open pipe, "
        "current state is PSInvocationState.Completed"
    )


def test_invalid_protocol_version():
    invalid = exp.InvalidProtocolVersion("reset runspace", PSVersion("1.2"), PSVersion("1.3"))
    assert str(invalid) == "reset runspace requires a protocol version of 1.3, current version is 1.2"
